//------------------------------------------------------------------------
// File for project XCALC
//-----------------------------------------------------------------------
// MATH functions
//-----------------------------------------------------------------------

#include "xcalc.h"
#include <time.h>

// Convenience declarations for angle measures
// Note: BC5.0 does not make proper LD values from decimal digits -
// therefore, we use bit patterns from BC3.1... (similar to M_PI_L)

#ifdef __BORLANDC__
char radsindeg[] = {0xae,0xc8,0xe9,0x94,0x12,0x35,0xfa,0x8e,0xf9,0x3f};
char degsinrad[] = {0xc3,0xbd,0x0f,0x1e,0xd3,0xe0,0x2e,0xe5,0x04,0x40};
#define DEGTORAD(x) (CURR_ANGLE==DEG?((x)**(LD*)radsindeg):(x))
#define RADTODEG(x) (CURR_ANGLE==DEG?((x)**(LD*)degsinrad):(x))
#else
// Skipping this for VC
#define DEGTORAD(x) (CURR_ANGLE==DEG?((x)*M_PI_L/180):(x))
#define RADTODEG(x) (CURR_ANGLE==DEG?((x)*180/M_PI_L):(x))
#endif

// This extern is not in XCALC.H
extern jmp_buf 	fatal_jmp;

//-----------------------------------------------------------------------
// Test functions for registers and numbers
//-----------------------------------------------------------------------
BOOL isint(Register *z)
{
	if ((z->ct==ctREAL || z->ct==ctDMS) && (z->imag!=0 || z->real>(LD)LONG_MAX || z->real<(LD)LONG_MIN || z->real!=(long)z->real)) return FALSE;
	/*
	31oct07: This raw cast to long fails under BCC for big numbers (not VC)
	if ((z->ct==ctREAL || z->ct==ctDMS) && (z->imag!=0 || z->real!=(long)z->real)) return FALSE;
	*/
	if (z->ct==ctFRACTION) return FALSE;
	return TRUE;
}

BOOL isproperfrac(Register *z)
{
	return z->ct==ctFRACTION;
}

BOOL isfrac(Register *z)
{
	return isproperfrac(z) || isint(z);
}

BOOL iseven(Register *z)
{
	return isint(z) && ((long)z->real&0x01)==0;
}

BOOL isreal(Register *z)
{
	return z->ct!=ctREAL || z->ct==ctREAL && z->imag==0.0;
}

BOOL iszero(Register *z)
{
	if ((z->ct==ctREAL || z->ct==ctDMS) && z->real==0.0 && z->imag==0.0) return TRUE;
	if (z->ct==ctDMS && z->real==0.0) return TRUE;
	if (z->ct==ctINTEGER && z->i==0) return TRUE;
	return FALSE;
}

BOOL ispositive(Register *z)
{
	if (z->ct==ctREAL || z->ct==ctDMS) return z->imag==0.0 && z->real>0.0;
	if (z->ct==ctFRACTION) return z->num>0;
	if (z->ct==ctINTEGER) return (z->i&0x80000000)==0;
	return FALSE;
}

BOOL isdms(Register *z)
{
	return z->ct==ctDMS;
}

BOOL islong(LD v)
{
	if (v>(LD)LONG_MAX) return FALSE;
	if (v<(LD)LONG_MIN) return FALSE;
	return v == (LD)(long)v;
}

BOOL islongrange(LD v)
{
	if (v>(LD)LONG_MAX) return FALSE;
	if (v<(LD)LONG_MIN) return FALSE;
	return TRUE;
}

BOOL candms(LD v)
{
	LD a = floorl(fabsl(v));
	if (a>(LD)DMSMAXINT) return FALSE;
	return TRUE;
}

void stickdms(Register *a, Register *b, Register *R)
{
	if ((isdms(a) || isdms(b)) && R->imag==0 && candms(R->real))
		R->ct = ctDMS;
	else
		R->ct = ctREAL;
}

//-----------------------------------------------------------------------
// set(c) sets global STR to c (only) for keyboard input
//-----------------------------------------------------------------------
void set(char c)
{
	memset(STR,0,NDIGITS+1);
	STR[0] = c;
}

//-----------------------------------------------------------------------
// app(c) appends c to global STR for keyboard input
//-----------------------------------------------------------------------
void app(char c)
{
	int l = lstrlen(STR);
	STR[l] = c;
	STR[l+1] = 0;
}

//-----------------------------------------------------------------------
// insbeg(c) inserts c left in global STR
//-----------------------------------------------------------------------
void insbeg(char c)
{
	int i,l;
	l=lstrlen(STR);
	for(i=l+1;i>0;i--) STR[i]=STR[i-1];
	STR[0]=c;
}

//-----------------------------------------------------------------------
// delbeg() clears first char in STR
//-----------------------------------------------------------------------
void delbeg(void)
{
	int i,l;
	l=lstrlen(STR);
	if(l==0) return;
	for(i=0;i<l;i++) STR[i]=STR[i+1];
}

//-----------------------------------------------------------------------
// Evaluate():
//		evaluate current input string STR using CURR_RADIX
//		delete rightmost character if an error (overflow and such) - return 1
//		update ST[0] with current value
//		reduce ST[0] if fraction
//		update display to show input string
//		return 0
//-----------------------------------------------------------------------
BOOL Evaluate(void)
{
	LD dval;
	int l,i;
	char *f1,*f2;

	l = lstrlen(STR);

	switch (CURR_RADIX) {
		case DECIMAL:
		{
			// First check for fractions (a/b/c)
			// (We can always set num/den to zero first)
			ST[Xreg].num = ST[Xreg].den = 0;
			if ((f1=strchr(STR,'/'))!=NULL) {
				long w;
				LD n,d;
				w = atol(STR);
				if (w==0) {	// Cannot have '0/' or '-0/'
					*f1 = 0;
					return 1;
				}
				n = _strtold(f1+1,NULL);
				if (n!=0) {
					if ((f2=strchr(f1+1,'/'))!=NULL) {
						d = _strtold(f2+1,NULL);
						if (d!=0) {
							LD ldnum = (LD)w*(LD)d + (w<0? -n : n);
							if (ldnum != (long)ldnum) {
								// number too big for fraction
								STR[l-1] = 0;
								return 1;
							}
							ST[Xreg].num = (long)(w*d + (w<0? -n : n));
							ST[Xreg].den = (long)d;
						}
						else {
							if (n != (long)n) {
								// number too big for fraction
								STR[l-1] = 0;
								return 1;
							}
							ST[Xreg].num = w;
							ST[Xreg].den = (long)n;
						}
					}
					else {
						if (n != (long)n) {
							// number too big for fraction
							STR[l-1] = 0;
							return 1;
						}
						ST[Xreg].num = w;
						ST[Xreg].den = (long)n;
					}
					ST[Xreg].real = ST[Xreg].imag = 0;
					Reduce(&ST[Xreg]);
				}
				else {
					ST[Xreg].real = (LD)w;
					ST[Xreg].imag = 0;
					if (*(f1+1)=='0' || *(f1+1)=='/') { // Cannot enter '/0' or '//'
						*(f1+1) = 0;
						return 1;
					}
				}
				// Reduced from fraction to real if num==0 (inelegant - maybe do this in Reduce())
				if (ST[Xreg].num) ST[Xreg].ct = ctFRACTION;
				break;
			}

			// Next check for DMS (dm's)
			if ((f1=strchr(STR,''))!=NULL) {
				int sign;
				long d,m;
				LD s=0;
				d = labs(atol(STR));
				if (!candms((LD)d)) {
					// d too large
					STR[l-1] = 0;
					return 1;
				}
				sign = *STR=='-'?-1:1;
				m = atol(f1+1);
				if (m>=60) {
					// no min>59
					STR[l-1] = 0;
					return 1;
				}
				if ((f2=strchr(f1+1,'\''))!=NULL) {
					s = _strtold(f2+1,NULL);
					if (s>=60) {
						// no sec>=60
						STR[l-1] = 0;
						return 1;
					}
				}
				ST[Xreg].real = d + m*sign/60.0L + s*sign/3600.0L;
				ST[Xreg].imag = 0;
				ST[Xreg].ct = ctDMS;
				break;
			}
			else if ((l==1)&&(STR[0]=='.') || (l==2)&&(STR[0]=='-')&&(STR[1]=='.')) {
				dval = 0;
				ST[Xreg].ct = ctREAL;
				break;
			}
			else if (sscanf(STR,"%Lg",&dval)!=1) {
				STR[l-1] = 0;
				return 1;
			}
			if (iState>=0) {
				ST[Xreg].real = dval;
				ST[Xreg].imag = 0;
			}
			else {
				ST[Xreg].imag = dval;
			}
			ST[Xreg].ct = ctREAL;
			break;
		}
		case COMPLEX:
		{
			if ((l==1)&&(STR[0]=='.') || (l==2)&&(STR[0]=='-')&&(STR[1]=='.')) {
				dval = 0;
				ST[Xreg].ct = ctREAL;
				break;
			}
			else if (sscanf(STR,"%Lg",&dval)!=1) {
				STR[l-1] = 0;
				return 1;
			}
			if (iState>=0) {
				ST[Xreg].real = dval;
				ST[Xreg].imag = 0;
			}
			else {
				ST[Xreg].imag = dval;
			}
			ST[Xreg].ct = ctREAL;
			break;
		}
		case HEX:
		{
			if (l>aMaxInputDigits[sWordSize][HEX-1]) {
				STR[l-1] = 0;
				return 1;
			}
			sscanf(STR,"%lx",&ST[Xreg].i);
			ST[Xreg].ct = ctINTEGER;
			break;
		}
		case OCTAL:
		{
			if (l>aMaxInputDigits[sWordSize][OCTAL-1]) {
				STR[l-1] = 0;
				return 1;
			}
			sscanf(STR,"%lo",&ST[Xreg].i);
			ST[Xreg].i &= CURR_MASK;
			sprintf(STR,"%lo",ST[Xreg].i);
			ST[Xreg].ct = ctINTEGER;
			break;
		}
		case BINARY:
		{
			if (l>aMaxInputDigits[sWordSize][BINARY-1]) {
				STR[l-1] = 0;
				return 1;
			}
			ST[Xreg].i=0;
			for(i=0;i<l;i++) {
				ST[Xreg].i<<=1;
				ST[Xreg].i+=(STR[i]-'0');
			}
			ST[Xreg].ct = ctINTEGER;
			break;
		}
	} // end switch CURR_RADIX
	return 0;
}

//-----------------------------------------------------------------------
// eps_check() nulls real or imaginary part if too small
//-----------------------------------------------------------------------

void eps_check(Register *r)
{
	LD re=r->real;
	LD im=r->imag;
	if(fabsl(re)<=LDBL_EPSILON*fabsl(im)) r->real=0.0;
	if(fabsl(im)<=LDBL_EPSILON*fabsl(re)) r->imag=0.0;
}

//-----------------------------------------------------------------------
// Reduce() reduces fraction to no common multipliers
//-----------------------------------------------------------------------
void Reduce(Register *x)
{
	long n, d, r;

	// We could end up here with a number such as "1/-2", so we check
	if (x->den < 0) {
		x->den = -x->den;
		x->num = -x->num;
	}

	n = x->num;
	d = x->den;

	// This routine is only called with proper fractions.
	// If a fractional result is zero, we will need to zero out the
	// real part; this will be used when num == 0.
	if (n == 0) {
		x->real = x->imag = 0.0;
		x->den = 0;
		x->ct = ctREAL;
		return; // No fraction
	}
	if (n<0) n = -n;

	while (d>0) {
		r = n%d;
		n = d;
		d = r;
	}

	x->num /= n;
	x->den /= n;

	if (x->den == 1) {
		x->real = x->num;
		x->imag = 0;
		x->num = x->den = 0;
		x->ct = ctREAL;
	}
	else {
		x->real = x->imag = 0.0;
	}
}

//-----------------------------------------------------------------------
// mkreal() converts fractional/DMS to real number
//-----------------------------------------------------------------------
void mkreal(Register *z)
{
	if (z->ct==ctFRACTION) {
		z->real = (LD)z->num / (LD)z->den;
		z->imag = 0;
		z->ct = ctREAL;
	}
	else if (z->ct==ctDMS) {
		z->ct = ctREAL;
	}
}

//-----------------------------------------------------------------------
// mkfrac() converts 'i n/d' number to 'n/d' (sets i to zero)
//-----------------------------------------------------------------------
void mkfrac(long *i,long *n,long *d)
{
	if (*i<0)
		*n = -*n + *i * *d;
	else
		*n += *i * *d;
	*i = 0;
}

//-----------------------------------------------------------------------
// mkifrac() converts 'n/d' number to 'i n/d' (takes out integer part)
// return value is 0 if number is positive, 1 if number is negative
//-----------------------------------------------------------------------
int mkifrac(long *i,long *n,long *d)
{
	int ret = 0;
	*i = *n / *d;
	if (*n < 0) ret = 1;
	*n = labs(*n % *d);
	return ret;
}

//-----------------------------------------------------------------------
// lift() lifts the stack
//-----------------------------------------------------------------------
void lift(void)
{
	int i;
	Dirty = TRUE;
	for (i=StkSiz-1;i;i--) {
		ST[i] = ST[i-1];
	}
}

//-----------------------------------------------------------------------
// drop() drops the stack
//-----------------------------------------------------------------------
void drop(void)
{
	int i;
	Dirty = TRUE;
	for (i=0;i<StkSiz-1;i++) {
		ST[i] = ST[i+1];
	}
	// New 18oct06: Option to not copy stack top on drop
	if (!CopyTop) {
		Register zero = {ctREAL,0};
		ST[StkSiz-1] = zero;
	}
}

//-----------------------------------------------------------------------
// clr(reg) clears one register
//-----------------------------------------------------------------------
void clrstkreg(int reg)
{
	clr(ST+reg);
}

//-----------------------------------------------------------------------
// clrreg(reg) clears one register
//-----------------------------------------------------------------------
void clr(Register *reg)
{
	reg->real = reg->imag = 0;
	reg->i = reg->num = reg->den = 0;
	reg->ct = ctREAL;
}

//-----------------------------------------------------------------------
// null(z) checks register z for underflow and sets to zero if so
//				 NB: we always set the register to ctREAL/ctDMS
//-----------------------------------------------------------------------
void null(Register *z)
{
	if (fabsl(z->real)<NUM_MIN) z->real = 0;
	if (fabsl(z->imag)<NUM_MIN) z->imag = 0;
	if (z->ct!=ctDMS) z->ct = ctREAL;
}

//-----------------------------------------------------------------------
// Lcheck(z) checks long double ld for overflow and returns 1 if so
//-----------------------------------------------------------------------
BOOL Lcheck(LD *val)
{
	return fabsl(*val)>(LD)NUM_MAX;
}

//-----------------------------------------------------------------------
// Rcheck(z) checks register z for overflow and returns 1 if so
//-----------------------------------------------------------------------
BOOL Rcheck(Register *z)
{
	return Lcheck(&z->real) || Lcheck(&z->imag);
}

//-----------------------------------------------------------------------
// _matherrl() - will resolve long double math library errors (not overflow)
//             and signal the calculator to display error message.
//-----------------------------------------------------------------------
#ifndef __BORLANDC__
#define _matherrl _matherr
#define _exceptionl _exception
#endif
int _matherrl(struct _exceptionl *e)
{
	e->retval = 0;
	switch (e->type) {
		case DOMAIN:	Error = __DOMAIN; break;
		case SING:		Error = __SINGULARITY; break;
		case OVERFLOW:	Error = __RANGE; break;
		case TLOSS:		Error = __TLOSS; break;
		default:		Error = __OTHER; break;
	}
	return 1;
}

//-----------------------------------------------------------------------
// FPE_Handler() - will resolve other FP math errors.
//             Currently we will have to terminate gracefully.
//-----------------------------------------------------------------------
void FPE_Handler(void)
{
	SetWindowPos(hwndXCALC,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
	DialogBox(hInst,"XCALCFATAL",hwndXCALC,FatalDlgProc);
	longjmp(fatal_jmp,1);
} 

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
// 							  Maths functions                          //
//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

//-----------------------------------------------------------------------
// Get functions - return fractional or real parts of register
// independent of representation (num always integer when this is called)
//-----------------------------------------------------------------------

LD getnum(Register *z)
{
	if (z->ct==ctFRACTION)
		return (LD)z->num;
	else
		return z->real;
}

LD getden(Register *z)
{
	if (z->ct==ctFRACTION)
		return (LD)z->den;
	else
		return 1.0L;
}

LD getreal(Register *z)
{
	if (z->ct==ctFRACTION)
		return (LD)z->num/(LD)z->den;
	else
		return z->real;
}

LD getimag(Register *z)
{
	if (z->ct==ctREAL)
		return z->imag;
	else
		return 0L;
}

//-----------------------------------------------------------------------
// Add() Y <- Y+X
// Return: zero=OK, nonzero=error
//-----------------------------------------------------------------------
int Add(void)
{
	LD re,im;
	if(wRadix==10) {
		if (isfrac(&ST[Xreg]) && isfrac(&ST[Yreg]) && (isproperfrac(ST+Xreg) || isproperfrac(ST+Yreg))) {
			LD n = getnum(&ST[Yreg])*getden(&ST[Xreg])+getnum(&ST[Xreg])*getden(&ST[Yreg]);
			LD d = getden(&ST[Yreg])*getden(&ST[Xreg]);
			if (islong(n) && islong(d)) {
				ST[Yreg].num = (long)n;
				ST[Yreg].den = (long)d;
				ST[Yreg].ct = ctFRACTION;
				Reduce(&ST[Yreg]);
				return 0;
			}
		}
		// Fall through if int overflow
		re=getreal(&ST[Yreg])+getreal(&ST[Xreg]);
		im=getimag(&ST[Yreg])+getimag(&ST[Xreg]);
		if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
		ST[Yreg].real=re;
		ST[Yreg].imag=im;
		null(&ST[Yreg]);  // zero if underflow (less than 1e-999). Always null frac.part.
		// ctDMS is sticky, if possible
		stickdms(ST+Xreg,ST+Yreg,ST+Yreg);
	}
	else {
		ST[Yreg].i+=ST[Xreg].i;
		ST[Yreg].i&=CURR_MASK;
	}
	return 0;
}

//-----------------------------------------------------------------------
// Sub() Y <- Y-X
//-----------------------------------------------------------------------
int Sub(void)
{
	LD re,im;
	if(wRadix==10) {
		if (isfrac(&ST[Xreg]) && isfrac(&ST[Yreg]) && (isproperfrac(ST+Xreg) || isproperfrac(ST+Yreg))) {
			LD n = getnum(&ST[Yreg])*getden(&ST[Xreg])-getnum(&ST[Xreg])*getden(&ST[Yreg]);
			LD d = getden(&ST[Yreg])*getden(&ST[Xreg]);
			if (islong(n) && islong(d)) {
				ST[Yreg].num = (long)n;
				ST[Yreg].den = (long)d;
				ST[Yreg].ct = ctFRACTION;
				Reduce(&ST[Yreg]);
				return 0;
			}
		}
		re=getreal(&ST[Yreg])-getreal(&ST[Xreg]);
		im=getimag(&ST[Yreg])-getimag(&ST[Xreg]);
		if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
		ST[Yreg].real=re;
		ST[Yreg].imag=im;
		null(&ST[Yreg]);
		// ctDMS is sticky, if possible
		stickdms(ST+Xreg,ST+Yreg,ST+Yreg);
	}
	else {
		ST[Yreg].i-=ST[Xreg].i;
		ST[Yreg].i&=CURR_MASK;
	}
	return 0;
}

//-----------------------------------------------------------------------
// Mul() Y <- Y*X
//-----------------------------------------------------------------------
int Mul(void)
{
	if(wRadix==10) {
		return ymulx(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
	}
	else {
		ST[Yreg].i*=ST[Xreg].i;
		ST[Yreg].i&=CURR_MASK;
	}
	return 0;
}

//-----------------------------------------------------------------------
// ymulx() multiplies two complex numbers
//-----------------------------------------------------------------------
int ymulx(Register *y,Register *x,Register *prod)
{
	LD re,im;    // long double
	if ((isproperfrac(y)||isproperfrac(x)) && isfrac(y) && isfrac(x)) {
		LD n = getnum(y)*getnum(x);
		LD d = getden(y)*getden(x);
		if (islong(n) && islong(d)) {
			prod->num = (long)n;
			prod->den = (long)d;
			prod->ct = ctFRACTION;
			Reduce(prod);		// Also clears real part
			return 0;
		}
	}
	re=getreal(y)*getreal(x)-getimag(y)*getimag(x);
	im=getreal(y)*getimag(x)+getimag(y)*getreal(x);
	if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
	prod->real = re;
	prod->imag = im;
	null(prod);
	// ctDMS is sticky, if possible
	stickdms(x,y,prod);
	return 0;
}

//-----------------------------------------------------------------------
// Div() Y <- Y/X
//-----------------------------------------------------------------------
int Div(void)
{
	if(wRadix==10) {
		return ydivx(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
	}
	else {
		if((ST[Xreg].i&CURR_MASK)==0) {
			return Error = __DIVBYZERO;
		}
		else {
			switch (sWordSize) {
				case 0:	ST[Yreg].i/=ST[Xreg].i; break; // 32 bits
				case 1:	ST[Yreg].i=(short)ST[Yreg].i/(short)ST[Xreg].i; break; // 16 bits
				case 2:	ST[Yreg].i=(char)ST[Yreg].i/(char)ST[Xreg].i; break; // 8 bits
			}
			ST[Yreg].i&=CURR_MASK;
		}
	}
	return 0;
}

//-----------------------------------------------------------------------
// ydivx() divides two complex numbers - used by several routines.
//-----------------------------------------------------------------------
int ydivx(Register *y,Register *x,Register *quot)
{
	LD re,im,denom;
	if ((isproperfrac(y)||isproperfrac(x)) && isfrac(y) && isfrac(x))
	{
		LD n = getnum(y)*getden(x);
		LD d = getden(y)*getnum(x);
		if (d==0) return Error = __DIVBYZERO;
		if (islong(n) && islong(d))
		{
			quot->num = (long)n;
			quot->den = (long)d;
			quot->ct = ctFRACTION;
			Reduce(quot);
			return 0;
		}
	}
	denom=getreal(x)*getreal(x)+getimag(x)*getimag(x);
	if (denom==(LD)0) return  Error = __DIVBYZERO;
	re=getreal(y)*getreal(x)+getimag(y)*getimag(x);
	im=getimag(y)*getreal(x)-getreal(y)*getimag(x);
	re/=denom;
	im/=denom;
	if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
	quot->real = re;
	quot->imag = im;
	null(quot);
	// ctDMS is sticky, if possible
	stickdms(x,y,quot);
	return 0;
}

//-----------------------------------------------------------------------
// Mod() Y <- Y%X (using INT in decimal/complex modes)
//-----------------------------------------------------------------------
int Mod(void)
{
	if(wRadix==10) {
		return ymodxf(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
	}
	else {
		if((ST[Xreg].i&CURR_MASK)==0) {
			return Error = __DIVBYZERO;
		}
		else {
			switch (sWordSize) {
				case 0: ST[Yreg].i%=ST[Xreg].i; break; // 32 bits
				case 1: ST[Yreg].i=(short)ST[Yreg].i%(short)ST[Xreg].i; break; // 16 bits
				case 2: ST[Yreg].i=(char)ST[Yreg].i%(char)ST[Xreg].i; break; // 8 bits
			}
			ST[Yreg].i&=CURR_MASK;
		}
	}
	return 0;
}

//-----------------------------------------------------------------------
// Divf() Y <- Y/X (using INT, for decimal/complex modes)
//-----------------------------------------------------------------------
int Divf(void)
{
	if(wRadix==10) {
		return ydivxf(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
	}
	return 0;
}

//-----------------------------------------------------------------------
// ymodxf() finds remainder of y/x, integer real parts only
//-----------------------------------------------------------------------
int ymodxf(Register *y,Register *x,Register *mod)
{
	LD rey,rex;
	long ly,lx;
	rey=getreal(y);
	rex=getreal(x);
	if (!islongrange(rey) || !islongrange(rex)) return Error = __RANGE;
	ly=(long)rey;
	lx=(long)rex;
	if (lx==0) return  Error = __DIVBYZERO;
	clr(mod);
	mod->real = ly%lx;
	return 0;
}

//-----------------------------------------------------------------------
// ydivxf() finds quotient of y/x, integer real parts only
//-----------------------------------------------------------------------
int ydivxf(Register *y,Register *x,Register *quot)
{
	LD rey,rex;
	long ly,lx;
	rey=getreal(y);
	rex=getreal(x);
	if (!islongrange(rey) || !islongrange(rex)) return Error = __RANGE;
	ly=(long)rey;
	lx=(long)rex;
	if (lx==0) return  Error = __DIVBYZERO;
	clr(quot);
	quot->real = ly/lx;
	return 0;
}

//-----------------------------------------------------------------------
// Percent() X <- Y*re(X)/100 (decimal modes only)
//-----------------------------------------------------------------------
int Percent(void)
{
	if ((isproperfrac(&ST[Yreg])||isproperfrac(&ST[Xreg])) && isfrac(&ST[Yreg]) && isfrac(&ST[Xreg]))
	{
		long n = (long)getnum(&ST[Yreg])*(long)getnum(&ST[Xreg]);
		long d = 100*(long)getden(&ST[Yreg])*(long)getden(&ST[Xreg]);
		ST[Xreg].real = ST[Xreg].imag = 0;
		ST[Xreg].num = n;
		ST[Xreg].den = d;
		ST[Xreg].ct = ctFRACTION;
		Reduce(&ST[Xreg]);
	}
	else
	{
		LD re,im;
		re=getreal(&ST[Yreg])*getreal(&ST[Xreg])/(LD)100.0L;
		im=getimag(&ST[Yreg])*getreal(&ST[Xreg])/(LD)100.0L;
		if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
		ST[Xreg].real=re;
		ST[Xreg].imag=im;
		null(&ST[Xreg]);
		stickdms(ST+Xreg,ST+Yreg,ST+Xreg);
	}
	return 0;
}

//-----------------------------------------------------------------------
void xchg(void)
{
	Register tmp=ST[Xreg];
	ST[Xreg]=ST[Yreg];
	ST[Yreg]=tmp;
}

int Sqr(void)
{
	return sqrx(&ST[Xreg],&ST[Xreg]);
}

int sqrx(Register *x,Register *square)
{
	if (wRadix==10) {
		int r = ymulx(x,x,square);
		if (r) return r;
		if (square->ct==ctDMS) square->ct=ctREAL;
		return 0;
	}
	else {
		square->i=x->i*x->i;
		square->i&=CURR_MASK;
	}
	return 0;
}

int Cube(void)
{
	return cubex(&ST[Xreg],&ST[Xreg]);
}

int cubex(Register *x,Register *cube)
{
	if (wRadix==10) {
		Register square;
		int r = ymulx(x,x,&square);
		if (r) return r;
		r = ymulx(x,&square,cube);
		if (r) return r;
		if (cube->ct==ctDMS) cube->ct=ctREAL;
		return 0;
	}
	else {
		cube->i=x->i*x->i*x->i;
		cube->i&=CURR_MASK;
	}
	return 0;
}

int Sqrt(void)
{
	sqrtx(&ST[Xreg],&ST[Xreg]);
	if(ST[Xreg].imag!=0.0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Qroot(void)
{
	Register tmp = {ctREAL,3,0};
	return yrootx(&ST[Xreg],&tmp,&ST[Xreg]);
}

int Rand(void)
{
	static BOOL first = TRUE;
	LD rand_num;
	if (first) {
		first = FALSE;
		srand((unsigned)time(NULL));
	}
	// Make random long double from successive calls to rand()
	rand_num = (LD)rand()/RAND_MAX
				+(LD)rand()/RAND_MAX/RAND_MAX
				+(LD)rand()/RAND_MAX/RAND_MAX/RAND_MAX
				+(LD)rand()/RAND_MAX/RAND_MAX/RAND_MAX/RAND_MAX;
	ST[Xreg].real = rand_num;
	ST[Xreg].imag = 0;
	ST[Xreg].ct = ctREAL;
	return 0;
}

// Cart() and Polar() behave differently in Decimal and Complex modes

int tocart(LD *x,LD *y)
{
	LD _x,_y;
	_x = *x * cosl(DEGTORAD(*y));
	if (Error) return Error;
	_y = *x * sinl(DEGTORAD(*y));
	if (Error) return Error;
	*x = _x;
	*y = _y;
	return 0;
}

int Cart(void)
{
	if (CURR_RADIX==COMPLEX)
		return tocart(&ST[Xreg].real,&ST[Xreg].imag);
	else
	{
		mkreal(&ST[Xreg]);
		mkreal(&ST[Yreg]);
		return tocart(&ST[Xreg].real,&ST[Yreg].real);
	}
}

int topolar(LD *r,LD *t)
{
	LD hyp,ang;
	hyp = hypotl(*r,*t);
	if (Error) return Error;
	ang = atan2l(*t,*r);
	if (Error) return Error;
	if (Lcheck(&hyp)) return Error = __RANGE;
	*r = hyp;
	*t = RADTODEG(ang);
	return 0;
}

int Polar(void)
{
	int r;
	if (CURR_RADIX==COMPLEX)
		return topolar(&ST[Xreg].real,&ST[Xreg].imag);
	else
	{
		mkreal(&ST[Xreg]);
		mkreal(&ST[Yreg]);
		r=topolar(&ST[Xreg].real,&ST[Yreg].real);
		if (!r) {
			// This is fun, but needs to check radians
			if (InvTrigAsDMS && CURR_ANGLE==DEG)
			ST[Yreg].ct = ctDMS;
		}
		return r;
	}
}

// Frac/Int calculates fractional/integer part of number

int Frac(void)
{
	if (isproperfrac(&ST[Xreg]))
	{
		ST[Xreg].num %= ST[Xreg].den;
	}
	else
	{
		int rsign = ST[Xreg].real<0?-1:1;
		int isign = ST[Xreg].imag<0?-1:1;
		ST[Xreg].real = (fabsl(ST[Xreg].real)-floorl(fabsl(ST[Xreg].real))) * rsign;
		ST[Xreg].imag = (fabsl(ST[Xreg].imag)-floorl(fabsl(ST[Xreg].imag))) * isign;
	}
	return 0;
}

int Int(void)
{
	if (isproperfrac(&ST[Xreg]))
	{
		ST[Xreg].num -= ST[Xreg].num % ST[Xreg].den;
		Reduce(&ST[Xreg]);
	}
	else
	{
		int rsign = ST[Xreg].real<0?-1:1;
		int isign = ST[Xreg].imag<0?-1:1;
		ST[Xreg].real = floorl(fabsl(ST[Xreg].real)) * rsign;
		ST[Xreg].imag = floorl(fabsl(ST[Xreg].imag)) * isign;
	}
	return 0;
}

int todms(Register *z)
{
	Register tmp = *z;
	if (isproperfrac(&tmp)) mkreal(&tmp);
	if (isreal(&tmp) && candms(tmp.real)) {
		tmp.ct = ctDMS;
		*z = tmp;
		return 0;
	}
	return 1;
}

int DMS(void)
{
	return todms(ST+Xreg);
}

void StripComma(char *s)
{
	char *p=s;
	do {
		if (*p==',') p++;
		*s++ = *p++;
	} while (*p);
	*s=0;
}

int Round(void)
{
	if (isproperfrac(&ST[Xreg])||isdms(&ST[Xreg])) {
		mkreal(&ST[Xreg]);
	}
	else if (isdms(&ST[Xreg])) {
		ST[Xreg].ct=ctREAL;
	}
	else {
		char Ntmp[80];
		LD real,imag=0;
		lstrcpy(Ntmp,cST[Xreg]);
		StripComma(Ntmp);
		if (CURR_RADIX==DECIMAL)
			sscanf(Ntmp,"%Lg",&real);
		else
			sscanf(Ntmp,"%Lg %Lg",&real,&imag);
		ST[Xreg].real = real;
		ST[Xreg].imag = imag;
	}
	return 0;
}

int sqrtx(Register *x,Register *y)
{
	Register tmp;
	if((getreal(x)==0.0)&&(getimag(x)==0.0)) {*y=*x; return 0;}
	lnx(x,&tmp);
	tmp.real/=2.0L;
	tmp.imag/=2.0L;
	return etox(&tmp,y); // no need to check overflow here
}

// It is a bit over-cautious to check overflow here, but then we don't
// need to check it elsewhere:

int etox(Register *x,Register *y)
{
	LD radius,re,im;
	radius=expl(getreal(x));
	if (Error) return Error;
	re=radius*cosl(getimag(x));
	im=radius*sinl(getimag(x));
	if (Error) return Error;
	if (Lcheck(&re) || Lcheck(&im)) return Error = __RANGE;
	y->real=re;
	y->imag=im;
	eps_check(y);
	null(y);
	y->ct=ctREAL;
	return 0;
}

int Exp(void)
{
	return etox(&ST[Xreg],&ST[Xreg]);
}

int lnx(Register *x,Register *y)
{
	LD re,im;
	re=hypotl(getimag(x),getreal(x));
	im=atan2l(getimag(x),getreal(x));
	re=logl(re);
	if (Error) return Error;
	y->real=re;
	y->imag=im;

	// This test is a bit questionable. It seems best to test the numbers
	// against each other, but we don't really know...
	//if(fabsl(re)<=LDBL_EPSILON*fabsl(im)) y->real=0.0;

	if(fabsl(re)<=LDBL_EPSILON) y->real=0.0;
	if(fabsl(im)<=LDBL_EPSILON*fabsl(re)) y->imag=0.0;

	null(y);
	y->ct=ctREAL;
	return 0;
}

int Ln(void)
{
	if (lnx(&ST[Xreg],&ST[Xreg])) return Error;
	if(ST[Xreg].imag!=0.0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Ten(void)
{
	Register tmp = {ctREAL,10,0};
	return ypowx(&tmp,&ST[Xreg],&ST[Xreg]);
}

int Log(void)
{
	Register tmp;
	if (lnx(&ST[Xreg],&tmp)) return Error;
	ST[Xreg].real=tmp.real/M_LN10;
	ST[Xreg].imag=tmp.imag/M_LN10;
	ST[Xreg].ct=ctREAL;
	null(&ST[Xreg]);
	if(ST[Xreg].imag!=0.0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

//-----------------------------------------------------------------------
// intpower() only called if y!=0 and x is integer
// Update 25/3/97: accept large negative power.
//-----------------------------------------------------------------------
int intpower(Register *y,long x,Register *result)
{
	Register prod = {ctREAL,1,0};	// initialise result to one
	Register zero = {ctREAL,0,0};
	Register temp = *y;				// temp for squaring where possible
	BOOL neg = (x<0);				// negative integer power
	if (temp.ct==ctDMS) temp.ct=ctREAL;			
	x = labs(x);
	
	// if power is 0, return 1
	if (!x) {
		*result = prod;
		return 0;
	}
	
	while (x) {
		// if odd power, multiply product with temp and decrease power by one
		if (x&1) {
			if (ymulx(&temp,&prod,&prod)) {
				if (neg) { // overflow & neg power -> zero. Must clear error
					*result = zero;
					return Error = __NOERROR;
				}
				else
					return Error = __RANGE;
			}
			x--;
		}
		if (!x) break;		// done
		// now power is even - square temp and half power
		if (ymulx(&temp,&temp,&temp)) {
			if (neg) { // overflow & neg power -> zero
				*result = zero;
				return Error = __NOERROR;
			}
			else
				return Error = __RANGE;
		}
		x>>=1;
	}
	if (!isfrac(&prod)) {
		null(&prod);
	}
	// result in prod. invert if negative power
	if (neg) {
		if (rcpx(&prod,result)) return Error = __SINGULARITY;
	}
	else {
		if (Rcheck(&prod)) return Error = __RANGE;
		*result = prod;
	}
	return 0;
}

int ypowx(Register *y,Register *x,Register *result)
{
	Register tmp = {ctREAL,0,0};
	LD re,im;
	if (!isint(x)) {
		if (iszero(y)) {
			if (ispositive(x)) {
				return 0;		// OK - zero to something real positive
			}
			else {
				// zero to none-positive number is singularity
				return Error = __SINGULARITY;
			}
		}
		if (lnx(y,&tmp)) return Error;
		re=tmp.real*getreal(x)-tmp.imag*getimag(x);
		im=tmp.real*getimag(x)+tmp.imag*getreal(x);
		tmp.real=re;
		tmp.imag=im;
		eps_check(&tmp);
		if (etox(&tmp,&tmp)) return Error;
		if (Rcheck(&tmp)) return Error = __RANGE;
		*result = tmp;
		if(result->imag!=0.0 && CURR_RADIX!=COMPLEX)
			SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
		return 0;
	}
	else {
		if (iszero(y)) {
			if (ispositive(x)) {
				return 0; 		// OK - zero to positive power is zero
			}
			else if (iszero(x)) {
				y->real = 1;
				y->imag = 0;
				null(y);
				return 0;		// OK - zero to zero is one
			}
			else {
				// zero to negative number is singularity
				return Error = __SINGULARITY;
			}
		}
		else {
			return intpower(y,(long)getreal(x),result);
		}
	}
}

//-----------------------------------------------------------------------
// Power() now tests if Xreg is integer, via ypowx()
//-----------------------------------------------------------------------
int Power(void)
{
	return ypowx(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
}

//-----------------------------------------------------------------------
// Root() now tests if Xreg is integer
//-----------------------------------------------------------------------
int yrootx(Register *y,Register *x,Register *result)
{
	Register tmp = {ctREAL,0,0};
	LD re,im,radius,angle;
	
	if (iszero(x))
		return Error = __SINGULARITY; // zero'th root is singularity 
	
	if (iszero(y)) {
		if (!ispositive(x)) return Error = __SINGULARITY; // zero to neg or complex is sing
		return 0; // zero to positive is zero
	}
	
	// integer root != zero and y is real: find correct root of y's radius, and set angle
	// depending on if x is odd (same angle) or even (divide angle by x)
	// 14jul05: odd roots never delivers complex results
	if (isint(x) && isreal(y)) {
		radius=fabsl(getreal(y));
		radius=powl(radius,1.0L/getreal(x));
		angle=atan2l(getimag(y),getreal(y));
		if (iseven(x)) {
			tmp.real = radius * cosl(angle/getreal(x));
			tmp.imag = radius * sinl(angle/getreal(x));
		}
		else {
			tmp.real = radius * cosl(angle);
			//tmp.imag = radius * sinl(angle);
		}
	}
	else {
		radius=getreal(x)*getreal(x)+getimag(x)*getimag(x);
		if (lnx(y,&tmp)) return Error;
		re=(tmp.real*getreal(x)+tmp.imag*getimag(x))/radius;
		im=(-tmp.real*getimag(x)+tmp.imag*getreal(x))/radius;
		tmp.real=re;
		tmp.imag=im;
		if (etox(&tmp,&tmp)) return Error;
	}
	
	if (Rcheck(&tmp)) return Error = __RANGE;
	
	*result = tmp;
	if(result->imag!=0.0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Root(void)
{
	return yrootx(&ST[Yreg],&ST[Xreg],&ST[Yreg]);
}

int Conjg(void)
{
	ST[Xreg].imag=(-ST[Xreg].imag);
	return 0;
}

int rcpx(Register *x,Register *quot)
{
	Register one = {ctREAL,1,0};
	return ydivx(&one,x,quot);
}

int Rcp(void)
{
	return rcpx(&ST[Xreg],&ST[Xreg]);
}

int Abs(void)
{
	if (isproperfrac(&ST[Xreg])) {
		ST[Xreg].num = labs(ST[Xreg].num);
	}
	else {
		LD re;
		re = hypotl(getreal(&ST[Xreg]),getimag(&ST[Xreg]));
		if (Error) return Error;
		if (Lcheck(&re)) return Error = __RANGE;
		ST[Xreg].real=re;
		ST[Xreg].imag=0.0;
		null(&ST[Xreg]);
	}
	return 0;
}

//-----------------------------------------------------------------------
// Math library coshl() and sinhl() from BC31 do not work correctly.
// Using my own _coshl() and _sinhl() instead.
//-----------------------------------------------------------------------
LD _coshl(LD z)
{
	return (expl(z)+expl(-z))/2.0L;
}

LD _sinhl(LD z)
{
	return (expl(z)-expl(-z))/2.0L;
}

int cosx(Register *x,Register *y)	// cos(re)cosh(im)-jsin(re)sinh(im)
{
	Register tmp1 = {ctREAL,0} ,tmp2 = {ctREAL,0};
	tmp1.real = DEGTORAD(getreal(x));
	tmp1.imag = DEGTORAD(getimag(x));
	tmp2.real = cosl(tmp1.real)*_coshl(tmp1.imag); // coshl generates error!!!
	tmp2.imag = -sinl(tmp1.real)*_sinhl(tmp1.imag);
	if (Error) return Error;
	eps_check(&tmp2);
	if (Rcheck(&tmp2)) return Error = __RANGE;
	y->real = tmp2.real;
	y->imag = tmp2.imag;
	y->ct = ctREAL;
	null(y);
	return 0;
}

int Cos(void)
{
	return cosx(&ST[Xreg],&ST[Xreg]);
}

int sinx(Register *x,Register *y)
{
	Register tmp1 = {ctREAL,0} ,tmp2 = {ctREAL,0};
	tmp1.real = DEGTORAD(getreal(x));
	tmp1.imag = DEGTORAD(getimag(x));
	tmp2.real = sinl(tmp1.real)*_coshl(tmp1.imag); // coshl generates error!!!
	tmp2.imag = cosl(tmp1.real)*_sinhl(tmp1.imag);
	if (Error) return Error;
	eps_check(&tmp2);
	if (Rcheck(&tmp2)) return Error = __RANGE;
	y->real = tmp2.real;
	y->imag = tmp2.imag;
	y->ct = ctREAL;
	null(y);
	return 0;
}

int Sin(void)
{
	return sinx(&ST[Xreg],&ST[Xreg]);
}

int tanx(Register *x,Register *y)
{
	Register tmp1,tmp2;
	cosx(x,&tmp2);
	sinx(x,&tmp1);
	if (Error) return Error;
	if (iszero(&tmp2))
		return Error = __SINGULARITY;
	else
		return ydivx(&tmp1,&tmp2,y);
}

int Tan(void)
{
	return tanx(&ST[Xreg],&ST[Xreg]);
}

int acosx(Register *x,Register *y)
{
	Register tmp1 = {ctREAL,0} ,tmp2 = {ctREAL,0};
	sqrx(x,&tmp1);
	if (Error) return Error;
	mkreal(&tmp1);
	tmp1.real-=1.0L;
	sqrtx(&tmp1,&tmp1);
	tmp1.real=getreal(x) - tmp1.real;
	tmp1.imag=getimag(x) - tmp1.imag;
	lnx(&tmp1,&tmp1);
	if (Error) return Error;
	// acos seems to return answer in forth quadrant (i.e. negative real part).
	// Try different approach:
	// tmp2.real=(tmp1.imag);
	// tmp2.imag=(-tmp1.real);
	tmp2.real=(-tmp1.imag);
	tmp2.imag=(-tmp1.real);
	eps_check(&tmp2);
	if (Rcheck(&tmp2)) return Error = __RANGE;
	if (tmp2.imag!=0.0) {
		*y=tmp2;
		if (CURR_RADIX!=COMPLEX) SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	}
	else {
		tmp2.real=RADTODEG(tmp2.real);
		if (Rcheck(&tmp2)) return Error = __RANGE;
		*y = tmp2;
	}
	null(y);
	// This is fun, but needs to check complex result or radians
	if (InvTrigAsDMS && CURR_ANGLE==DEG && CURR_RADIX!=COMPLEX)
		y->ct = ctDMS;
	return 0;
}

int Acos(void)
{
	return acosx(&ST[Xreg],&ST[Xreg]);
}

int asinx(Register *x,Register *y)
{
	Register tmp1 = {ctREAL,0} ,tmp2 = {ctREAL,0};
	sqrx(x,&tmp1);
	if (Error) return Error;
	mkreal(&tmp1);
	tmp1.real=1.0L-tmp1.real;
	tmp1.imag=(-tmp1.imag);
	sqrtx(&tmp1,&tmp1);
	tmp2.real=tmp1.real + getimag(x);
	tmp2.imag=tmp1.imag - getreal(x);
	lnx(&tmp2,&tmp2);
	if (Error) return Error;
	// asin seems to return answer with wrong principal imag value.
	// Try different approach:
	// tmp1.real=(-tmp2.imag);
	// tmp1.imag=(tmp2.real);
	tmp1.real=(-tmp2.imag);
	tmp1.imag=(-tmp2.real);
	eps_check(&tmp1);
	if (Rcheck(&tmp1)) return Error = __RANGE;
	if (tmp1.imag!=0) {
		*y=tmp1;
		if (CURR_RADIX!=COMPLEX) SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	}
	else {
		tmp1.real=RADTODEG(tmp1.real);
		if (Rcheck(&tmp1)) return Error = __RANGE;
		*y = tmp1;
	}
	null(y);
	// This is fun, but needs to check complex result or radians
	if (InvTrigAsDMS && CURR_ANGLE==DEG && CURR_RADIX!=COMPLEX)
		y->ct = ctDMS;
	return 0;
}

int Asin(void)
{
	return asinx(&ST[Xreg],&ST[Xreg]);
}

int atanx(Register *x,Register *y)
{
	// To avoid rounding problems (resulting in complex numbers),
	// we here make real calculations when possible...

	Register tmp1 = {ctREAL,0} ,tmp2 = {ctREAL,0};

	if (isreal(x)) {
		tmp2.real = atanl(getreal(x));
		if (Error) return Error;
	}
	else {
		tmp2.real=(-(tmp1.real=getreal(x)));
		tmp1.imag=getimag(x)+1.0L;
		tmp2.imag=1.0L-getimag(x);
		ydivx(&tmp1,&tmp2,&tmp1);
		if (Error) return Error;
		lnx(&tmp1,&tmp1);
		if (Error) return Error;
		tmp2.real=(-tmp1.imag)/2.0L;
		tmp2.imag=tmp1.real/2.0L;
	}

	eps_check(&tmp2);
	if (Rcheck(&tmp2)) return Error = __RANGE;
	if (tmp2.imag!=0.0) {
		*y = tmp2;
		if (CURR_RADIX!=COMPLEX) SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	}
	else {
		tmp2.real=RADTODEG(tmp2.real);
		if (Rcheck(&tmp2)) return Error = __RANGE;
		*y = tmp2;
	}
	null(y);
	// This is fun, but needs to check complex result or radians
	if (InvTrigAsDMS && CURR_ANGLE==DEG && CURR_RADIX!=COMPLEX)
		y->ct = ctDMS;
	return 0;
}

int Atan(void)
{
	return atanx(&ST[Xreg],&ST[Xreg]);
}

int coshx(Register *x,Register *y)
{
	Register p={ctREAL,0}, m={ctREAL,0};
	p.real = getreal(x);
	p.imag = getimag(x);
	m.real = -getreal(x);
	m.imag = -getimag(x);
	etox(&p,&p);
	etox(&m,&m);
	if (Error) return Error;
	p.real = (p.real+m.real)/2;
	p.imag = (p.imag+m.imag)/2;
	if (Rcheck(&p)) return Error = __RANGE;
	*y = p;
	null(y);
	return 0;
}

int Cosh(void)
{
  return coshx(&ST[Xreg],&ST[Xreg]);
}

int sinhx(Register *x,Register *y)
{
	Register p={ctREAL,0}, m={ctREAL,0};
	p.real = getreal(x);
	p.imag = getimag(x);
	m.real = -getreal(x);
	m.imag = -getimag(x);
	etox(&p,&p);
	etox(&m,&m);
	if (Error) return Error;
	p.real = (p.real-m.real)/2;
	p.imag = (p.imag-m.imag)/2;
	if (Rcheck(&p)) return Error = __RANGE;
	*y = p;
	null(y);
	return 0;
}

int Sinh(void)
{
  return sinhx(&ST[Xreg],&ST[Xreg]);
}

int tanhx(Register *x,Register *y)
{
	Register tmp1,tmp2;
	coshx(x,&tmp2);
	sinhx(x,&tmp1);
	if (Error) return Error;
	return ydivx(&tmp1,&tmp2,y);
}

int Tanh(void)
{
	return tanhx(&ST[Xreg],&ST[Xreg]);
}

int acoshx(Register *x,Register *y)
{
	Register tmp;
	sqrx(x,&tmp);
	if (Error) return Error;
	mkreal(&tmp);
	tmp.real -= 1.0L;
	sqrtx(&tmp,&tmp);
	if (Error) return Error;
	tmp.real += getreal(x);
	tmp.imag += getimag(x);
	lnx(&tmp,y);
	if (Error) return Error;
	eps_check(y);
	//y->ct = ctREAL;
	if (y->imag!=0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Acosh(void)
{
  return acoshx(&ST[Xreg],&ST[Xreg]);
}

int asinhx(Register *x,Register *y)
{
	Register tmp;
	sqrx(x,&tmp);
	if (Error) return Error;
	mkreal(&tmp);
	tmp.real += 1.0L;
	sqrtx(&tmp,&tmp);
	if (Error) return Error;
	tmp.real += getreal(x);
	tmp.imag += getimag(x);
	lnx(&tmp,y);
	if (Error) return Error;
	eps_check(y);
	//y->ct = ctREAL;
	if (y->imag!=0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Asinh(void)
{
  return asinhx(&ST[Xreg],&ST[Xreg]);
}

int atanhx(Register *x,Register *y)
{
	Register tmp1={ctREAL,0}, tmp2={ctREAL,0};
	tmp1.real = 1.0L + getreal(x);
	tmp1.imag = getimag(x);
	tmp2.real = 1.0L - getreal(x);
	tmp2.imag = -getimag(x);
	ydivx(&tmp1,&tmp2,&tmp1);
	if (Error) return Error;
	sqrtx(&tmp1,&tmp1);
	if (Error) return Error;
	lnx(&tmp1,y);
	if (Error) return Error;
	eps_check(y);
	//y->ct = ctREAL;
	if (y->imag!=0 && CURR_RADIX!=COMPLEX)
		SendMessage(hwndXCALC,PM_SETMODE,COMPLEX,0);
	return 0;
}

int Atanh(void)
{
	return atanhx(&ST[Xreg],&ST[Xreg]);
}

void tofloat(void)
{
	int i;
	for(i=0;i<StkSiz;i++) {
		ST[i].real=(LD)(ST[i].i|(((ST[i].i>>(CURR_BITS-1))&0x1)*~CURR_MASK));
		ST[i].imag=0.0;
		ST[i].num=ST[i].den=0;
		ST[i].ct=ctREAL;
	}
	ST[Lreg].real=(LD)(ST[Lreg].i|(((ST[Lreg].i>>(CURR_BITS-1))&0x1)*~CURR_MASK));
	ST[Lreg].imag=0.0;
	ST[Lreg].num=ST[Lreg].den=0;
	ST[Lreg].ct=ctREAL;
}

int Fact(void)
{
	LD tmp;
	int i;
	tmp = floorl(getreal(&ST[Xreg]));
	if ((tmp>FACT_MAX)||(tmp<0)) return Error = __RANGE;
	ST[Xreg].imag = 0.0;
	ST[Xreg].real = 1;
	for (i=(int)tmp;i>0;i--) ST[Xreg].real *= (LD)i;
	null(&ST[Xreg]);
	ST[Xreg].ct = ctREAL;
	return 0;
}

void toint(void)
{
	int i;
	LD v;
	if (Error) return;
	// Silently ignore overflow in LastX
	if (!ZeroOverflow) for (i=0;i<StkSiz;i++) {
		v=getreal(ST+i);
		// Round towards zero to mimic long cast
		if (v<0)
			v = ceill(v);
		else
			v = floorl(v);
		Error|=(v>CURR_MAX||v<CURR_MIN);
	}
	if (Error) {
		Error = __RANGE;
	}
	else {
		for (i=0;i<StkSiz;i++) {
			v = getreal(ST+i);
			if (v<0)
				v = ceill(v);
			else
				v = floorl(v);
			ST[i].i = ((v>CURR_MAX||v<CURR_MIN)?0:(long)v) & CURR_MASK;
			ST[i].ct = ctINTEGER;
		}
		v = getreal(ST+Lreg);
		if (v<0)
			v = ceill(v);
		else
			v = floorl(v);
		ST[Lreg].i = ((v>CURR_MAX||v<CURR_MIN)?0:(long)v) & CURR_MASK;
		ST[Lreg].ct = ctINTEGER;
	}
}

void rolldn(void)
{
	int i;
	Register tmp=ST[Xreg];
	for(i=Xreg;i<StkSiz-1;i++) ST[i]=ST[i+1];
	ST[StkSiz-1]=tmp;
}

void rollup(void)
{
	Register tmp=ST[StkSiz-1];
	lift();
	ST[Xreg]=tmp;
}

void X__Y(void)
{
	Register tmp = ST[Xreg];
	ST[Xreg] = ST[Yreg];
	ST[Yreg] = tmp;
}

void R__I(void)
{
	LD tmp = ST[Xreg].real;
	ST[Xreg].real = ST[Xreg].imag;
	ST[Xreg].imag = tmp;
}

void CHS(void)
{
	ST[Xreg].real = -ST[Xreg].real;
	ST[Xreg].imag = -ST[Xreg].imag;
	ST[Xreg].i = (-ST[Xreg].i)&CURR_MASK;
	ST[Xreg].num = -ST[Xreg].num;
}

int OR(void)
{
	ST[Yreg].i |= ST[Xreg].i;
	return 0;
}

int AND(void)
{
	ST[Yreg].i &= ST[Xreg].i;
	return 0;
}

int XOR(void)
{
	ST[Yreg].i ^= ST[Xreg].i;
	return 0;
}

int NOT(void)
{
	ST[Xreg].i = (~ST[Xreg].i)&CURR_MASK;
	return 0;
}

int LShift(void)
{
	ST[Xreg].i = (ST[Xreg].i<<1)&CURR_MASK;
	return 0;
}

int LRot(void)
{
	ST[Xreg].i = (long)((ST[Xreg].i<<1)|((unsigned long)ST[Xreg].i>>(CURR_BITS-1)))&CURR_MASK;
	return 0;
}

int LAShift(void)
{
	ST[Xreg].i = (ST[Xreg].i<<1)&CURR_MASK;
	return 0;
}

int RShift(void)
{
	ST[Xreg].i = (long)((unsigned long)ST[Xreg].i>>1)&CURR_MASK;
	return 0;
}

int RRot(void)
{
	ST[Xreg].i = (long)(((unsigned long)ST[Xreg].i>>1)|((unsigned long)ST[Xreg].i<<(CURR_BITS-1)))&CURR_MASK;
	return 0;
}

int RAShift(void)
{
	ST[Xreg].i = ((ST[Xreg].i|~CURR_MASK)>>1)&CURR_MASK;
	return 0;
}

