//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// ToolTip event handler and window procedure
// Actions:
//	- handle all mouse events to provide tooltips in any window
//	- manage and draw tooltip window
//------------------------------------------------------------------------

#include "xcalc.h"

static BOOL bToolTipShowing = FALSE;
static BOOL bToolTipCanShow = TRUE;
static BOOL bToolTipFollow = FALSE;
static HWND hwndCapture = NULL;
static char sToolTipText[TOOLTIP_MAXLENGTH];

/*
 * Tooltip window always exists, but is shown only when mouse is resting on a
 * window that provides text when ToolTip option is on.
 */
void ToolTipEvent(HWND hwnd,UINT wMsg,WPARAM wParam,LPARAM lParam,RECT *prect,const char *text)
{
	POINTS pts;
	POINT pt;
	BOOL bInside;
	BOOL bButtonDown;
	HWND hwndTmp;
	BOOL bFocus = FALSE;
	char *p;

	// Shall we show tooltips?
	if (!ToolTip)
		return;
	// Do we have focus?
	hwndTmp = GetActiveWindow();
	while (hwndTmp) {
		if (hwndTmp == hwndXCALC) {
			bFocus = TRUE;
			break;
		}
		hwndTmp = GetParent(hwndTmp);
	}
	if (!bFocus)
		return;

	switch (wMsg) {
	case WM_MOUSEMOVE:
		pts = MAKEPOINTS(lParam);
		pt.x = pts.x;
		pt.y = pts.y;
		bInside = PtInRect(prect,pt);
		bButtonDown = wParam & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON);

		if (!bButtonDown) {
			if (bInside) {
				if (!hwndCapture) {
					SetCapture(hwnd);
					//MessageBeep(MB_ICONEXCLAMATION);
					hwndCapture = hwnd;
					lstrcpyn(sToolTipText,text,sizeof(sToolTipText)-1); // need room for double 0 at end
					// Break into lines and count them
					p = sToolTipText;
					while (*p) {
						if (*p == '\n') {
							*p = 0;
						}
						p++;
					}
					*++p = 0;
					//ToolTipCanShow = TRUE;
				}
				SetTimer(hwndToolTip,TOOLTIP_SHOWTIMER,
					bToolTipFollow?TOOLTIP_FOLLOWSHOWMS:TOOLTIP_SHOWMS,NULL);
			}
			if (hwndCapture && !bInside) {
				KillTimer(hwndToolTip,TOOLTIP_SHOWTIMER);
				ReleaseCapture();
				//MessageBeep(MB_ICONASTERISK);
				hwndCapture = NULL;
				ShowWindow(hwndToolTip,SW_HIDE);
				if (bToolTipShowing) {
					bToolTipFollow = TRUE;
				}
				bToolTipShowing = FALSE;
				bToolTipCanShow = TRUE;
				SetTimer(hwndToolTip,TOOLTIP_FOLLOWTIMER,TOOLTIP_FOLLOWMS,NULL);
			}
		}
		break;
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_MBUTTONDOWN:
		HideToolTip();
		break;
	}
}

void HideToolTip()
{
	KillTimer(hwndToolTip,TOOLTIP_SHOWTIMER);
	KillTimer(hwndToolTip,TOOLTIP_HIDETIMER);
	KillTimer(hwndToolTip,TOOLTIP_FOLLOWTIMER);
	bToolTipCanShow = FALSE;
	bToolTipFollow = FALSE;
	if (hwndCapture) {
		ReleaseCapture();
		hwndCapture = NULL;
	}
	if (bToolTipShowing) {
		ShowWindow(hwndToolTip,SW_HIDE);
		bToolTipShowing = FALSE;
	}
}

LRESULT CALLBACK ToolTipProc(HWND hwnd, UINT wMsg, WPARAM wParam, LPARAM lParam)
{
	static DWORD txtCol;
	static int textDy;
	HDC hdc;
	HPEN hpen;
	HGDIOBJ oldfont;
	HGDIOBJ oldpen;
	PAINTSTRUCT ps;
	RECT rect;
	char *p;
	int outDy;

	switch (wMsg) {
	case WM_TIMER:
		if (wParam == TOOLTIP_SHOWTIMER) {
			bToolTipFollow = FALSE;
			KillTimer(hwnd,TOOLTIP_SHOWTIMER);
			if (!bToolTipShowing && bToolTipCanShow && lstrlen(sToolTipText)) {
				POINT pt;
				if (GetCursorPos(&pt)) {
					SIZE siz;
					RECT rect;
					int height = 0;
					int width = 0;
					txtCol = GetSysColor(COLOR_INFOTEXT);
					hdc = GetDC(hwnd);
					oldfont = SelectObject(hdc,hToolTipFont);
					p = sToolTipText;
					while (*p) {
						GetTextExtentPoint32(hdc,p,lstrlen(p),&siz);
						if (siz.cx > width) width = siz.cx;
						textDy = siz.cy;
						height += siz.cy;
						p += lstrlen(p)+1;
					}
					SelectObject(hdc,oldfont);
					ReleaseDC(hwnd,hdc);
					// Rect here used as (x0,y0,w,h)
					rect.left = pt.x;
					rect.top = pt.y + GetSystemMetrics(SM_CYCURSOR)/2+3;
					rect.right = width+6;
					rect.bottom = height+4;
					siz.cx = GetSystemMetrics(SM_CXSCREEN);
					siz.cy = GetSystemMetrics(SM_CYSCREEN);
					if (rect.left+rect.right>siz.cx) rect.left = siz.cx - rect.right;
					if (rect.top+rect.bottom>siz.cy) rect.top = siz.cy - rect.bottom;
					SetWindowPos(hwnd,HWND_TOPMOST,rect.left,rect.top,rect.right,rect.bottom,
						SWP_SHOWWINDOW|SWP_NOACTIVATE);
					bToolTipShowing = TRUE;
					bToolTipCanShow = FALSE;
					SetTimer(hwndToolTip,TOOLTIP_HIDETIMER,TOOLTIP_HIDEMS,NULL);
				}
			}
		}
		else if (wParam == TOOLTIP_HIDETIMER) {
			bToolTipFollow = FALSE;
			KillTimer(hwnd,TOOLTIP_HIDETIMER);
			if (bToolTipShowing) {
				ShowWindow(hwndToolTip,SW_HIDE);
				bToolTipShowing = FALSE;
				bToolTipCanShow = FALSE;
			}
		}
		else if (wParam == TOOLTIP_FOLLOWTIMER) {
			bToolTipFollow = FALSE;
			KillTimer(hwnd,TOOLTIP_FOLLOWTIMER);
		}
		break;

	case WM_PAINT:
		BeginPaint(hwnd,&ps);
		hpen = CreatePen(PS_SOLID,1,(COLORREF)txtCol);
		oldfont = SelectObject(ps.hdc,hToolTipFont);
		oldpen = SelectObject(ps.hdc,hpen);
		GetClientRect(hwnd,&rect);
		Rectangle(ps.hdc,rect.left++,rect.top++,rect.right--,rect.bottom--);
		FillRect(ps.hdc, &rect,(HBRUSH)(COLOR_INFOBK+1));
		SetBkMode(ps.hdc,TRANSPARENT);
		SetTextColor(ps.hdc,(COLORREF)txtCol);
		p = sToolTipText;
		outDy = 0;
		while (*p) {
			TextOut(ps.hdc,3,2+outDy,p,lstrlen(p));
			outDy += textDy;
			p += lstrlen(p)+1;
		}
		SelectObject(ps.hdc,oldfont);
		SelectObject(ps.hdc,oldpen);
		DeleteObject(hpen);
		EndPaint(hwnd,&ps);
		break;

	default:
		return DefWindowProc(hwnd,wMsg,wParam,lParam);
	}
	return 0;
}
