//------------------------------------------------------------------------
// File for project XCALC
//------------------------------------------------------------------------
// Keyboard procedure
// Actions:
//	- process keypresses and send messages to main window procedure
//	- update display to show keypress
//  - manage right-click help menu
//------------------------------------------------------------------------

#include "xcalc.h"

// Special care needed to put an UNDO button here - we cannot simulate
// Alt-Backspace with keycodes, since it seems Windows catches this
// combination. We must therefore send an accelerator command AS_UNDO
// for the Undo key:
// 7/1/97 This goes even if ^Z also does undo. We cannot simulate an accel key.

// 14/3/97: To accomodate any accelerator sent from the keyboard, a new
// and more general approach is taken, using constant _AS as a flag.

// 4/7/97: Added support for hyperbolic functions. These require
// special visual feedback, since we must share the same buttons as
// for the normal trig functons. Sorry about that.

// 21/7/00: ToolTips added.

// 16/3/08: Better internal repr. of keys by using struct 'key'

// #defines for allowed modes for a key (for editing)
#define NOEDIT		(0)
#define ALL			(DECIMAL|COMPLEX|HEX|OCTAL|BINARY)
#define DECCPX		(DECIMAL|COMPLEX)
#define BINMODES	(HEX|OCTAL|BINARY)
#define NOTBINOCT	(DECIMAL|COMPLEX|HEX)
#define NOTBINARY	(DECIMAL|COMPLEX|HEX|OCTAL)

// Global variables used in editing keys (Dialogs.c)

int KeyToEdit;
int ModeToEdit;
int WindowToEdit;
HWND HwndToEdit;
RECT RectToEdit;

// All defined keys (Note: additions at end - otherwise existing keyboards are corrupted)
key Key[NKEYID] = {
	/*ID			Prog name	Mode		Caption		Sym	Operation				Tooltip */
	{KEY_NONE,		"NONE",		ALL,		"",			-1,	{0,0},					"(Empty)"},
	{KEY_0,			"0",		ALL,		"0",		-1,	{0,'0'},				"0"},
	{KEY_1,			"1",		ALL,		"1",		-1,	{0,'1'},				"1"},
	{KEY_2,			"2",		NOTBINARY,	"2",		-1,	{0,'2'},				"2"},
	{KEY_3,			"3",		NOTBINARY,	"3",		-1,	{0,'3'},				"3"},
	{KEY_4,			"4",		NOTBINARY,	"4",		-1,	{0,'4'},				"4"},
	{KEY_5,			"5",		NOTBINARY,	"5",		-1,	{0,'5'},				"5"},
	{KEY_6,			"6",		NOTBINARY,	"6",		-1,	{0,'6'},				"6"},
	{KEY_7,			"7",		NOTBINARY,	"7",		-1,	{0,'7'},				"7"},
	{KEY_8,			"8",		NOTBINOCT,	"8",		-1,	{0,'8'},				"8"},
	{KEY_9,			"9",		NOTBINOCT,	"9",		-1,	{0,'9'},				"9"},
	{KEY_A,			"A",		HEX,		"A",		-1,	{0,'a'},				"A [a]"},
	{KEY_B,			"B",		HEX,		"B",		-1,	{0,'b'},				"B [b]"},
	{KEY_C,			"C",		HEX,		"C",		-1,	{0,'c'},				"C [c]"},
	{KEY_D,			"D",		HEX,		"D",		-1,	{0,'d'},				"D [d]"},
	{KEY_E,			"E",		HEX,		"E",		-1,	{0,'e'},				"E [e]"},
	{KEY_F,			"F",		HEX,		"F",		-1,	{0,'f'},				"F [f]"},
	{KEY_POINTP,	".",		DECCPX,		".",		-1,	{0,'.'},				"Decimal [. (point)]"},
	{KEY_POINTC,	"",			NOEDIT,		".",		-1,	{0,','},				"Decimal [, (comma)]"},
	{KEY_CHS,		"CHS",		ALL,		"+/-",		 2,	{0,'m'},				"Change sign [m]"},
	{KEY_EEX,		"EEX",		DECCPX,		"E",		-1,	{0,'e'},				"Enter exponent [e]"},
	{KEY_BACK,		"BACK",		ALL,		"",		 0,	{0,8},					"Backspace/clear X [Backspace]"},
	{KEY_ENTER,		"ENTER",	ALL,		"Enter",	-1,	{0,13},					"Enter/push stack"},
	{KEY_ADD,		"+",		ALL,		"+",		-1,	{0,'+'},				"Addition [+]"},
	{KEY_SUB,		"-",		ALL,		"-",		 0,	{0,'-'},				"Subtraction [-]"},
	{KEY_MUL,		"*",		ALL,		"",		 0,	{0,'*'},				"Multiplication [*]"},
	{KEY_DIV,		"/",		ALL,		"",		 0,	{0,'/'},				"Division [/]"},
	{KEY_DIVF,		"DIV",		DECCPX,		"div",		-1,	{0,'D'},				"Integer division [Shift-D]"},
	{KEY_MOD,		"MOD",		BINMODES,	"%",		-1,	{0,'%'},				"Modulo [%]"},
	{KEY_MODF,		"MOD",		DECCPX,		"mod",		-1,	{0,'M'},				"Modulo [Shift-M]"},
	{KEY_LASTX,		"LASTX",	ALL,		"LastX",	-1,	{0,'l'},				"Last X [lower l]"},
	{KEY_UNDO,		"",			ALL,		"Undo",		-1,	{_AS,AS_UNDO},			"Undo"},
	{KEY_CLSTK,		"CLSTK",	ALL,		"Clstk",	-1,	{0,'Z'},				"Clear stack [Shift-Z]"},
	{KEY_IJ,		"IJ",		DECCPX,		"y+ix",		-1,	{0,'i'},				"Y, X to complex [i]"},
	{KEY_FRAC,		"FRAC",		DECIMAL,	"a b/c",	-1,	{0,' '},				"Enter fraction [f or <space>]"},
	{KEY_DMS,		"DMS",		DECIMAL,	" ' \"",	-1,	{0,'\''},				"Enter degree minute second [']"},
	{KEY_TODMS,		"TODMS",	DECIMAL,	" ' \"",	 0,	{0,4},					"To degree minute second [Ctrl-D]"},
	{KEY_ROUND,		"ROUND",	DECCPX,		"Rnd",		-1,	{0,'$'},				"Round [$ or Shift-R]"},
	{KEY_XY,		"XY",		ALL,		"xy",		 1,	{0,'<'},				"Swap X and Y [< or >]"},
	{KEY_RUP,		"RUP",		ALL,		"",		 0,	{VK_UP,0},				"Roll up [Up]"},
	{KEY_RDN,		"RDN",		ALL,		"",		 0,	{VK_DOWN,0},			"Roll down [Down]"},
	{KEY_PERC,		"%",		DECCPX,		"%",		-1,	{0,'%'},				"Percent [%]"},
	{KEY_CONV,		"",			DECCPX,		"Conv",		-1,	{0,'v'},				"Conversions [v]"},
	{KEY_QCONV,		"QCONV",	DECCPX,		"Qcnv",		-1,	{0,'V'},				"Quick convert [Shift-V]"},
	{KEY_CONS,		"",			DECCPX,		"Cons",		-1,	{0,'k'},				"Constants [k]"},
	{KEY_QCONS,		"QCONS",	DECCPX,		"Qcns",		-1,	{0,'K'},				"Quick constant [Shift-K]"},
	{KEY_RAND,		"RAND",		DECCPX,		"Rand",		-1,	{0,'?'},				"Pseudo random number [?]"},
	{KEY_TOFRAC,	"TOFRAC",	DECCPX,		"Frac",		-1,	{0,6},					"Fractional part [Ctrl-F]"},
	{KEY_TOINT,		"TOINT",	DECCPX,		"Int",		-1,	{0,9},					"Integer part [Ctrl-I]"},
	{KEY_CART,		"CART",		DECCPX,		"x:y",		-1,	{0,':'},				"Cartesian [: (colon)]"},
	{KEY_POLAR,		"POLAR",	DECCPX,		"r;f",		 2,	{0,';'},				"Polar [; (semicolon)]"},
	{KEY_EXP,		"EXP",		DECCPX,		"exp",		-1,	{0,'E'},				"Exponential [Shift-E]"},
	{KEY_TEN,		"TEN",		DECCPX,		"10^x",		-1,	{0,'N'},				"10's exponent [Shift-N]"},
	{KEY_ROOT,		"ROOT",		DECCPX,		"xy",		 1,	{0,'&'},				"Root [r or &]"},
	{KEY_SQRT,		"SQRT",		DECCPX,		"x",		 0,	{0,'q'},				"Square root [q]"},
	{KEY_QROOT,		"QROOT",	DECCPX,		"x",		 1,	{0,'u'},				"Cube root [u]"},
	{KEY_LN,		"LN",		DECCPX,		"ln",		-1,	{0,'L'},				"Natural logarithm [Shift-L]"},
	{KEY_LOG,		"LOG",		DECCPX,		"log",		-1,	{0,'G'},				"Logarithm [Shift-G]"},
	{KEY_POW,		"POW",		DECCPX,		"y^x",		-1,	{0,'^'},				"Power [o or ^]"},
	{KEY_SQR,		"SQR",		DECCPX,		"x",		-1,	{0,'Q'},				"Square [Shift-Q]"},
	{KEY_CUBE,		"CUBE",		DECCPX,		"x",		-1,	{0,'U'},				"Cube [Shift-U]"},
	{KEY_ABS,		"ABS",		DECCPX,		"|x|",		-1,	{0,'|'},				"Absolute value [|]"},
	{KEY_SIN,		"SIN",		DECCPX,		"sin",		-1,	{0,'s'},				"Sine [s]"},
	{KEY_COS,		"COS",		DECCPX,		"cos",		-1,	{0,'c'},				"Cosine [c]"},
	{KEY_TAN,		"TAN",		DECCPX,		"tan",		-1,	{0,'t'},				"Tangent [t]"},
	{KEY_ASIN,		"ASIN",		DECCPX,		"asin",		-1,	{0,'S'},				"Arcsine [Shift-S]"},
	{KEY_ACOS,		"ACOS",		DECCPX,		"acos",		-1,	{0,'C'},				"Arccosine [Shift-C]"},
	{KEY_ATAN,		"ATAN",		DECCPX,		"atan",		-1,	{0,'T'},				"Arctangent [Shift-T]"},
	{KEY_SINH,		"SINH",		NOEDIT,		"sinh",		-1,	{0,'s'},				"Sine hyperbolicus [s]"},
	{KEY_COSH,		"COSH",		NOEDIT,		"cosh",		-1,	{0,'c'},				"Cosine hyperbolicus [c]"},
	{KEY_TANH,		"TANH",		NOEDIT,		"tanh",		-1,	{0,'t'},				"Tangent hyperbolicus [t]"},
	{KEY_ASINH,		"ASINH",	NOEDIT,		"asinh",	-1,	{0,'S'},				"Arsine hyperbolicus [Shift-S]"},
	{KEY_ACOSH,		"ACOSH",	NOEDIT,		"acosh",	-1,	{0,'C'},				"Arcosine hyperbolicus [Shift-C]"},
	{KEY_ATANH,		"ATANH",	NOEDIT,		"atanh",	-1,	{0,'T'},				"Artangent hyperbolicus [Shift-T]"},
	{KEY_HYP,		"",			DECCPX,		"hyp",		-1,	{0,'h'},				"Hyperbolic functions [h]"},
	{KEY_NONHYP,	"",			NOEDIT,		"hyp",		-1,	{0,'H'},				"Non-hyperbolic functions [Shift-H]"},
	{KEY_RCP,		"RCP",		DECCPX,		"1/x",		-1,	{0,'\\'},				"Reciprocal [\\]"},
	{KEY_PI,		"PI",		DECCPX,		"p",		 0,	{0,'p'},				"Pi [p]"},
	{KEY_FACT,		"FACT",		DECCPX,		"x!",		-1,	{0,'!'},				"Factorial [!]"},
	{KEY_IMAGP,		"",			NOEDIT,		"",		 0,	{0,'.'},				"Enter imaginary part [. (point)]"},
	{KEY_IMAGC,		",",		COMPLEX,	"",		 0,	{0,','},				"Enter imaginary part [, (comma)]"},
	{KEY_NIJ,		"NIJ",		COMPLEX,	"y,x",		 0,	{0,'I'},				"Complex to Y,X [Shift-I]"},
	{KEY_REIM,		"REIM",		COMPLEX,	"RI",		 1,	{0,'\"'},				"Swap real and imaginary [\"]"},
	{KEY_CONJ,		"CONJ",		COMPLEX,	"Conj",		-1,	{0,'\''},				"Conjugate [']"},
	{KEY_AND,		"AND",		BINMODES,	"And",		-1,	{0,'A'},				"And [Shift-A or &]"},
	{KEY_OR,		"OR",		BINMODES,	"Or",		-1,	{0,'o'},				"Or [o or |]"},
	{KEY_XOR,		"XOR",		BINMODES,	"Xor",		-1,	{0,'x'},				"Xor [x or ^]"},
	{KEY_NOT,		"NOT",		BINMODES,	"Not",		-1,	{0,'n'},				"Not [n or ~]"},
	{KEY_SHL,		"SHL",		BINMODES,	"Sh<",		-1,	{VK_LEFT,0},			"Logical shift left [Left]"},
	{KEY_SHR,		"SHR",		BINMODES,	"Sh>",		-1,	{VK_RIGHT,0},			"Logical shift right [Right]"},
	{KEY_ASHL,		"ASHL",		BINMODES,	"Ash<",		-1,	{VK_SHIFT,VK_LEFT},		"Arithmetic shift left [Shift-left]"},
	{KEY_ASHR,		"ASHR",		BINMODES,	"Ash>",		-1,	{VK_SHIFT,VK_RIGHT},	"Arithmetic shift right [Shift-right]"},
	{KEY_ROTL,		"ROTL",		BINMODES,	"Rot<",		-1,	{_AS,AS_CTRL_LEFT},		"Rotate left [Ctrl-left]"},
	{KEY_ROTR,		"ROTR",		BINMODES,	"Rot>",		-1,	{_AS,AS_CTRL_RIGHT},	"Rotate right [Ctrl-right]"},
	{KEY_STO0,		"STO0",		ALL,		"Sto 0",	-1,	{_AS,AS_STORE0},		"Store register 0 [Ctrl-numeric-0]"},
	{KEY_STO1,		"STO1",		ALL,		"Sto 1",	-1,	{_AS,AS_STORE1},		"Store register 1 [Ctrl-numeric-1]"},
	{KEY_STO2,		"STO2",		ALL,		"Sto 2",	-1,	{_AS,AS_STORE2},		"Store register 2 [Ctrl-numeric-2]"},
	{KEY_STO3,		"STO3",		ALL,		"Sto 3",	-1,	{_AS,AS_STORE3},		"Store register 3 [Ctrl-numeric-3]"},
	{KEY_STO4,		"STO4",		ALL,		"Sto 4",	-1,	{_AS,AS_STORE4},		"Store register 4 [Ctrl-numeric-4]"},
	{KEY_STO5,		"STO5",		ALL,		"Sto 5",	-1,	{_AS,AS_STORE5},		"Store register 5 [Ctrl-numeric-5]"},
	{KEY_STO6,		"STO6",		ALL,		"Sto 6",	-1,	{_AS,AS_STORE6},		"Store register 6 [Ctrl-numeric-6]"},
	{KEY_STO7,		"STO7",		ALL,		"Sto 7",	-1,	{_AS,AS_STORE7},		"Store register 7 [Ctrl-numeric-7]"},
	{KEY_STO8,		"STO8",		ALL,		"Sto 8",	-1,	{_AS,AS_STORE8},		"Store register 8 [Ctrl-numeric-8]"},
	{KEY_STO9,		"STO9",		ALL,		"Sto 9",	-1,	{_AS,AS_STORE9},		"Store register 9 [Ctrl-numeric-9]"},
	{KEY_RCL0,		"RCL0",		ALL,		"Rcl 0",	-1,	{_AS,AS_RECALL0},		"Recall register 0 [Alt-numeric-0]"},
	{KEY_RCL1,		"RCL1",		ALL,		"Rcl 1",	-1,	{_AS,AS_RECALL1},		"Recall register 1 [Alt-numeric-1]"},
	{KEY_RCL2,		"RCL2",		ALL,		"Rcl 2",	-1,	{_AS,AS_RECALL2},		"Recall register 2 [Alt-numeric-2]"},
	{KEY_RCL3,		"RCL3",		ALL,		"Rcl 3",	-1,	{_AS,AS_RECALL3},		"Recall register 3 [Alt-numeric-3]"},
	{KEY_RCL4,		"RCL4",		ALL,		"Rcl 4",	-1,	{_AS,AS_RECALL4},		"Recall register 4 [Alt-numeric-4]"},
	{KEY_RCL5,		"RCL5",		ALL,		"Rcl 5",	-1,	{_AS,AS_RECALL5},		"Recall register 5 [Alt-numeric-5]"},
	{KEY_RCL6,		"RCL6",		ALL,		"Rcl 6",	-1,	{_AS,AS_RECALL6},		"Recall register 6 [Alt-numeric-6]"},
	{KEY_RCL7,		"RCL7",		ALL,		"Rcl 7",	-1,	{_AS,AS_RECALL7},		"Recall register 7 [Alt-numeric-7]"},
	{KEY_RCL8,		"RCL8",		ALL,		"Rcl 8",	-1,	{_AS,AS_RECALL8},		"Recall register 8 [Alt-numeric-8]"},
	{KEY_RCL9,		"RCL9",		ALL,		"Rcl 9",	-1,	{_AS,AS_RECALL9},		"Recall register 9 [Alt-numeric-9]"},
	{KEY_CLR0,		"CLR0",		ALL,		"Clr 0",	-1,	{_AS,AS_CLEAR0},		"Clear register 0 [Ctrl-Alt-numeric-0]"},
	{KEY_CLR1,		"CLR1",		ALL,		"Clr 1",	-1,	{_AS,AS_CLEAR1},		"Clear register 1 [Ctrl-Alt-numeric-1]"},
	{KEY_CLR2,		"CLR2",		ALL,		"Clr 2",	-1,	{_AS,AS_CLEAR2},		"Clear register 2 [Ctrl-Alt-numeric-2]"},
	{KEY_CLR3,		"CLR3",		ALL,		"Clr 3",	-1,	{_AS,AS_CLEAR3},		"Clear register 3 [Ctrl-Alt-numeric-3]"},
	{KEY_CLR4,		"CLR4",		ALL,		"Clr 4",	-1,	{_AS,AS_CLEAR4},		"Clear register 4 [Ctrl-Alt-numeric-4]"},
	{KEY_CLR5,		"CLR5",		ALL,		"Clr 5",	-1,	{_AS,AS_CLEAR5},		"Clear register 5 [Ctrl-Alt-numeric-5]"},
	{KEY_CLR6,		"CLR6",		ALL,		"Clr 6",	-1,	{_AS,AS_CLEAR6},		"Clear register 6 [Ctrl-Alt-numeric-6]"},
	{KEY_CLR7,		"CLR7",		ALL,		"Clr 7",	-1,	{_AS,AS_CLEAR7},		"Clear register 7 [Ctrl-Alt-numeric-7]"},
	{KEY_CLR8,		"CLR8",		ALL,		"Clr 8",	-1,	{_AS,AS_CLEAR8},		"Clear register 8 [Ctrl-Alt-numeric-8]"},
	{KEY_CLR9,		"CLR9",		ALL,		"Clr 9",	-1,	{_AS,AS_CLEAR9},		"Clear register 9 [Ctrl-Alt-numeric-9]"},
	{KEY_FIX,		"FIX",		DECCPX,		"Fix",		-1,	{_AS,AS_FIX},			"Fix [Alt-I]"},
	{KEY_SCI,		"SCI",		DECCPX,		"Sci",		-1,	{_AS,AS_SCI},			"Scientific [Alt-S]"},
	{KEY_ENG,		"ENG",		DECCPX,		"Eng",		-1,	{_AS,AS_ENG},			"Engineering [Alt-E]"},
	{KEY_FLOAT,		"FLOAT",	DECCPX,		"Float",	-1,	{_AS,AS_FLOAT},			"Float [Alt-F]"},
	{KEY_FIX0,		"FIX0",		DECCPX,		"Fix 0",	-1,	{_AS,AS_FIX0},			"Fix 0 [Alt-0]"},
	{KEY_FIX1,		"FIX1",		DECCPX,		"Fix 1",	-1,	{_AS,AS_FIX1},			"Fix 1 [Alt-1]"},
	{KEY_FIX2,		"FIX2",		DECCPX,		"Fix 2",	-1,	{_AS,AS_FIX2},			"Fix 2 [Alt-2]"},
	{KEY_FIX3,		"FIX3",		DECCPX,		"Fix 3",	-1,	{_AS,AS_FIX3},			"Fix 3 [Alt-3]"},
	{KEY_FIX4,		"FIX4",		DECCPX,		"Fix 4",	-1,	{_AS,AS_FIX4},			"Fix 4 [Alt-4]"},
	{KEY_FIX5,		"FIX5",		DECCPX,		"Fix 5",	-1,	{_AS,AS_FIX5},			"Fix 5 [Alt-5]"},
	{KEY_FIX6,		"FIX6",		DECCPX,		"Fix 6",	-1,	{_AS,AS_FIX6},			"Fix 6 [Alt-6]"},
	{KEY_FIX7,		"FIX7",		DECCPX,		"Fix 7",	-1,	{_AS,AS_FIX7},			"Fix 7 [Alt-7]"},
	{KEY_FIX8,		"FIX8",		DECCPX,		"Fix 8",	-1,	{_AS,AS_FIX8},			"Fix 8 [Alt-8]"},
	{KEY_FIX9,		"FIX9",		DECCPX,		"Fix 9",	-1,	{_AS,AS_FIX9},			"Fix 9 [Alt-9]"},
	{KEY_DEG,		"DEG",		DECIMAL,	"Deg",		-1,	{_AS,AS_DEG},			"Degree [Alt-G]"},
	{KEY_RAD,		"RAD",		DECIMAL,	"Rad",		-1,	{_AS,AS_RAD},			"Radian [Alt-R]"},
	{KEY_DECIMAL,	"DECIMAL",	ALL,		"Dec",		-1,	{_AS,AS_DECIMAL},		"Decimal [Alt-D]"},
	{KEY_COMPLEX,	"COMPLEX",	ALL,		"Cpx",		-1,	{_AS,AS_COMPLEX},		"Complex [Alt-C]"},
	{KEY_HEX,		"HEX",		ALL,		"Hex",		-1,	{_AS,AS_HEX},			"Hex [Alt-H]"},
	{KEY_OCTAL,		"OCTAL",	ALL,		"Oct",		-1,	{_AS,AS_OCTAL},			"Octal [Alt-O]"},
	{KEY_BINARY,	"BINARY",	ALL,		"Bin",		-1,	{_AS,AS_BINARY},		"Binary [Alt-B]"},
	{KEY_32BIT,		"32BIT",	BINMODES,	"32bit",	-1,	{_AS,AS_32BIT},			"32 bit words [Alt-3]"},
	{KEY_16BIT,		"16BIT",	BINMODES,	"16bit",	-1,	{_AS,AS_16BIT},			"16 bit words [Alt-1]"},
	{KEY_8BIT,		"8BIT",		BINMODES,	"8bit",		-1,	{_AS,AS_8BIT},			"8 bit words [Alt-8]"},
	{KEY_CONFIG,	"",			ALL,		"Cfg",		-1,	{_AS,AS_CONFIG},		"Configuration [Alt-N]"},
	{KEY_EXIT,		"EXIT",		ALL,		"Exit",		-1,	{_AS,AS_EXIT},			"Exit [Alt-X]"},
	{KEY_HELP,		"",			ALL,		"Help",		-1,	{_AS,AS_HELP},			"Help [F1]"},
	{KEY_KEYHELP,	"",			ALL,		"Keyh",		-1,	{_AS,AS_KEYHELP},		"Keyboard help [Ctrl-F1]"},
	{KEY_COPY,		"COPY",		ALL,		"Copy",		-1,	{_AS,AS_COPY},			"Copy [Ctrl-C]"},
	{KEY_PASTE,		"PASTE",	ALL,		"Paste",	-1,	{_AS,AS_PASTE},			"Paste [Ctrl-V]"},
	{KEY_POPUP,		"",			ALL,		"Mem",		-1,	{_AS,AS_POPUP},			"Toggle memory popup [Ctrl-N]"},
	{KEY_MINI,		"",			ALL,		"Mini",		-1,	{_AS,AS_MINI},			"Toggle mini/normal [Alt-M]"},
	{KEY_PROG1,		"",			ALL,		"P1",		-1,	{_AS,AS_PROG1},			"Run program 1 [Shift-F1]"},
	{KEY_PROG2,		"",			ALL,		"P2",		-1,	{_AS,AS_PROG2},			"Run program 2 [Shift-F2]"},
	{KEY_PROG3,		"",			ALL,		"P3",		-1,	{_AS,AS_PROG3},			"Run program 3 [Shift-F3]"},
	{KEY_PROG4,		"",			ALL,		"P4",		-1,	{_AS,AS_PROG4},			"Run program 4 [Shift-F4]"},
	{KEY_PROG5,		"",			ALL,		"P5",		-1,	{_AS,AS_PROG5},			"Run program 5 [Shift-F5]"},
	{KEY_PROG6,		"",			ALL,		"P6",		-1,	{_AS,AS_PROG6},			"Run program 6 [Shift-F6]"},
	{KEY_PROG7,		"",			ALL,		"P7",		-1,	{_AS,AS_PROG7},			"Run program 7 [Shift-F7]"},
	{KEY_PROG8,		"",			ALL,		"P8",		-1,	{_AS,AS_PROG8},			"Run program 8 [Shift-F8]"},
	{KEY_PROG9,		"",			ALL,		"P9",		-1,	{_AS,AS_PROG9},			"Run program 9 [Shift-F9]"},
	{KEY_PROG10,	"",			ALL,		"P10",		-1,	{_AS,AS_PROG10},		"Run program 10 [Shift-F10]"},
	{KEY_PROG11,	"",			ALL,		"P11",		-1,	{_AS,AS_PROG11},		"Run program 11 [Shift-F11]"},
	{KEY_PROG12,	"",			ALL,		"P12",		-1,	{_AS,AS_PROG12},		"Run program 12 [Shift-F12]"},
	{KEY_EDITPR,	"",			ALL,		"Prog",		-1,	{_AS,AS_EDITPROGRAM},	"Edit programs [Alt-F1]"},
	{KEY_CLRMEM,	"CLRMEM",	ALL,		"Clr M",	-1,	{_AS,AS_CLEARMEM},		"Clear all memory registers [Ctrl-Alt-M]"},
	{KEY_FIXP,		"FIX+",		DECCPX,		"Fix+",		-1,	{VK_LEFT,0},			"Increment FIX [Left]"},
	{KEY_FIXM,		"FIX-",		DECCPX,		"Fix-",		-1,	{VK_RIGHT,0},			"Decrement FIX [Right]"},
};

// Default key mappings.

#define NDFLTKEYS 66
const int DfltKeys6x11[NRADIX][NDFLTKEYS] = {
	// Decimal
	{
		KEY_PROG1,	KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_CART,	KEY_ROUND,	KEY_TODMS,	KEY_DMS,	KEY_FRAC,	KEY_PERC,	
		KEY_PROG2,	KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_POLAR,	KEY_TOFRAC,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_PROG3,	KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_RAND,	KEY_TOINT,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_PROG4,	KEY_HYP,	KEY_FACT,	KEY_SQR,	KEY_SQRT,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_PROG5,	KEY_CONV,	KEY_CONS,	KEY_RCP,	KEY_QROOT,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_PROG6,	KEY_QCONV,	KEY_QCONS,	KEY_PI,		KEY_ABS,	KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Complex
	{
		KEY_PROG1,	KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_CART,	KEY_ROUND,	KEY_IMAGC,	KEY_TOFRAC,	KEY_TOINT,	KEY_PERC,	
		KEY_PROG2,	KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_POLAR,	KEY_CONJ,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_PROG3,	KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_REIM,	KEY_NIJ,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_PROG4,	KEY_HYP,	KEY_FACT,	KEY_SQR,	KEY_SQRT,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_PROG5,	KEY_CONV,	KEY_CONS,	KEY_RCP,	KEY_QROOT,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_PROG6,	KEY_QCONV,	KEY_QCONS,	KEY_PI,		KEY_ABS,	KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Hex
	{
		KEY_PROG1,	0,			0,			KEY_AND,	0,			0,			KEY_C,		KEY_D,		KEY_E,		KEY_F,		KEY_MOD,	
		KEY_PROG2,	KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			KEY_8,		KEY_9,		KEY_A,		KEY_B,		KEY_ADD,	
		KEY_PROG3,	KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_SUB,	
		KEY_PROG4,	KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_PROG5,	0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		KEY_PROG6,	0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Octal
	{
		KEY_PROG1,	0,			0,			KEY_AND,	0,			0,			0,			0,			0,			0,			KEY_MOD,	
		KEY_PROG2,	KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_ADD,	
		KEY_PROG3,	KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_SUB,	
		KEY_PROG4,	KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		0,			0,			0,			0,			KEY_MUL,	
		KEY_PROG5,	0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		KEY_PROG6,	0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Binary
	{
		KEY_PROG1,	0,			0,			KEY_AND,	0,			0,			0,			0,			0,			0,			KEY_MOD,	
		KEY_PROG2,	KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			0,			0,			0,			0,			KEY_ADD,	
		KEY_PROG3,	KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			0,			KEY_0,		KEY_1,		0,			KEY_SUB,	
		KEY_PROG4,	KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		0,			0,			0,			0,			KEY_MUL,	
		KEY_PROG5,	0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		KEY_PROG6,	0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
};

const int DfltKeys6x10[NRADIX][NDFLTKEYS] = {
	// Decimal
	{
		KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_CART,	KEY_ROUND,	KEY_TODMS,	KEY_DMS,	KEY_FRAC,	KEY_PERC,	
		KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_POLAR,	KEY_TOFRAC,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_RAND,	KEY_TOINT,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_HYP,	KEY_FACT,	KEY_SQR,	KEY_SQRT,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_CONV,	KEY_CONS,	KEY_RCP,	KEY_QROOT,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_QCONV,	KEY_QCONS,	KEY_PI,		KEY_ABS,	KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Complex
	{
		KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_CART,	KEY_ROUND,	KEY_IMAGC,	KEY_TOFRAC,	KEY_TOINT,	KEY_PERC,	
		KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_POLAR,	KEY_CONJ,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_REIM,	KEY_NIJ,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_HYP,	KEY_FACT,	KEY_SQR,	KEY_SQRT,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_CONV,	KEY_CONS,	KEY_RCP,	KEY_QROOT,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_QCONV,	KEY_QCONS,	KEY_PI,		KEY_ABS,	KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Hex
	{
		0,			0,			KEY_AND,	0,			0,			KEY_C,		KEY_D,		KEY_E,		KEY_F,		KEY_MOD,	
		KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			KEY_8,		KEY_9,		KEY_A,		KEY_B,		KEY_ADD,	
		KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_SUB,	
		KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Octal
	{
		0,			0,			KEY_AND,	0,			0,			0,			0,			0,			0,			KEY_MOD,	
		KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_ADD,	
		KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_SUB,	
		KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		0,			0,			0,			0,			KEY_MUL,	
		0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Binary
	{
		0,			0,			KEY_AND,	0,			0,			0,			0,			0,			0,			KEY_MOD,	
		KEY_SHL,	KEY_SHR,	KEY_OR,		0,			0,			0,			0,			0,			0,			KEY_ADD,	
		KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			0,			0,			KEY_0,		KEY_1,		0,			KEY_SUB,	
		KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_XY,		0,			0,			0,			0,			KEY_MUL,	
		0,			0,			0,			0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			KEY_DIV,	
		0,			0,			0,			0,			KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
};

const int DfltKeys5x12[NRADIX][NDFLTKEYS] = {
	// Decimal
	{
		KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_RAND,	KEY_PERC,	KEY_TODMS,	KEY_DMS,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_TOFRAC,	KEY_CONV,	KEY_ROUND,	KEY_FRAC,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_TOINT,	KEY_QCONV,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_HYP,	KEY_RCP,	KEY_SQR,	KEY_SQRT,	KEY_CART,	KEY_CONS,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_PI,		KEY_FACT,	KEY_ABS,	KEY_QROOT,	KEY_POLAR,	KEY_QCONS,	KEY_RDN,	KEY_UNDO,	KEY_LASTX,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Complex
	{
		KEY_SIN,	KEY_ASIN,	KEY_LN,		KEY_EXP,	KEY_ROUND,	KEY_PERC,	KEY_CONJ,	KEY_LASTX,	KEY_7,		KEY_8,		KEY_9,		KEY_ADD,	
		KEY_COS,	KEY_ACOS,	KEY_LOG,	KEY_TEN,	KEY_TOFRAC,	KEY_CONV,	KEY_REIM,	KEY_NIJ,	KEY_4,		KEY_5,		KEY_6,		KEY_SUB,	
		KEY_TAN,	KEY_ATAN,	KEY_POW,	KEY_ROOT,	KEY_TOINT,	KEY_QCONV,	KEY_XY,		KEY_IJ,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		KEY_HYP,	KEY_RCP,	KEY_SQR,	KEY_SQRT,	KEY_CART,	KEY_CONS,	KEY_RUP,	KEY_CLSTK,	KEY_EEX,	KEY_0,		KEY_POINTP,	KEY_DIV,	
		KEY_PI,		KEY_FACT,	KEY_ABS,	KEY_QROOT,	KEY_POLAR,	KEY_QCONS,	KEY_RDN,	KEY_UNDO,	KEY_IMAGC,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	
	},
	// Hex
	{
		0,			0,			0,			0,			0,			0,			0,			KEY_C,		KEY_D,		KEY_E,		KEY_F,		KEY_ADD,	
		0,			0,			0,			KEY_AND,	0,			0,			0,			KEY_8,		KEY_9,		KEY_A,		KEY_B,		KEY_SUB,	
		0,			KEY_SHL,	KEY_SHR,	KEY_OR,		0,			KEY_XY,		0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_MUL,	
		0,			KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			KEY_RUP,	KEY_CLSTK,	KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_DIV,	
		0,			KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_RDN,	KEY_LASTX,	KEY_UNDO,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	KEY_MOD,
	},
	// Octal
	{
		0,			0,			0,			0,			0,			0,			0,			0,			0,			0,			0,			KEY_ADD,	
		0,			0,			0,			KEY_AND,	0,			0,			0,			KEY_4,		KEY_5,		KEY_6,		KEY_7,		KEY_SUB,	
		0,			KEY_SHL,	KEY_SHR,	KEY_OR,		0,			KEY_XY,		0,			KEY_0,		KEY_1,		KEY_2,		KEY_3,		KEY_MUL,	
		0,			KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			0,			KEY_DIV,	
		0,			KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_RDN,	KEY_LASTX,	KEY_UNDO,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	KEY_MOD,
	},
	// Binary
	{
		0,			0,			0,			0,			0,			0,			0,			0,			0,			0,			0,			KEY_ADD,	
		0,			0,			0,			KEY_AND,	0,			0,			0,			0,			0,			0,			0,			KEY_SUB,	
		0,			KEY_SHL,	KEY_SHR,	KEY_OR,		0,			KEY_XY,		0,			0,			KEY_0,		KEY_1,		0,			KEY_MUL,	
		0,			KEY_ASHL,	KEY_ASHR,	KEY_XOR,	0,			KEY_RUP,	KEY_CLSTK,	0,			0,			0,			0,			KEY_DIV,	
		0,			KEY_ROTL,	KEY_ROTR,	KEY_NOT,	0,			KEY_RDN,	KEY_LASTX,	KEY_UNDO,	KEY_CHS,	KEY_BACK,	KEY_ENTER,	KEY_MOD,
	},
};

int *Keys[NRADIX] = {NULL};

int sameKeys(int a, int b)
{
	if (a>b) {int t=a; a=b; b=t;}
	return a==b || a==KEY_POINTP&&b==KEY_POINTC || a==KEY_IMAGP&&b==KEY_IMAGC;
}

void *dfltKeys(void)
{
	const int (*dflt)[NRADIX][NDFLTKEYS] = NULL;
	if (nKeyCols==10) dflt = &DfltKeys6x10;
	if (nKeyCols==11) dflt = &DfltKeys6x11;
	if (nKeyCols==12) dflt = &DfltKeys5x12;
	return (void*)dflt;
}

void SetDefaultKeyboard(int mode)
{
	int k;
	const int (*dflt)[NRADIX][NDFLTKEYS] = dfltKeys();
	if (dflt) {
		for (k=0;k<nKeyRows*nKeyCols;k++) {
			Keys[mode][k] = k<NDFLTKEYS?(*dflt)[mode][k]:0;
		}
	}
}

BOOL initKeys(void)
{
	int m,k;
	// This tests that all keys are in order
	for (k=0;k<NKEYID;k++) {
		if (Key[k].id!=k) {
			MessageBox(NULL,"Wrong KEYID. Fix and recompile.","XCALC internal error",MB_OK);
			return FALSE;
		}
	}
	freeKeys();
	for (m=0;m<NRADIX;m++) {
		Keys[m] = calloc(nKeyRows*nKeyCols,sizeof(int));
		SetDefaultKeyboard(m);
	}
	return TRUE;
}

void freeKeys(void)
{
	int m;
	for (m=0;m<NRADIX;m++) {
		if (Keys[m]) {
			free(Keys[m]);
			Keys[m] = NULL;
		}
	}
}

// #defines to apply keyboard fudges
#define DECIMALSET 0
#define COMPLEXSET 1

// This routine is a fudge to shift the keys "." and "," if we use
// decimal COMMA

void KeyboardDecimal(BOOL useComma)
{
	int i;
	if (useComma) {
		for (i=0;i<nKeyRows*nKeyCols;i++) {
			if (Keys[DECIMALSET][i]==KEY_POINTP) Keys[DECIMALSET][i]=KEY_POINTC;
			if (Keys[COMPLEXSET][i]==KEY_POINTP) Keys[COMPLEXSET][i]=KEY_POINTC;
			if (Keys[COMPLEXSET][i]==KEY_IMAGC)  Keys[COMPLEXSET][i]=KEY_IMAGP;
		}
	} else {
		for (i=0;i<nKeyRows*nKeyCols;i++) {
			if (Keys[DECIMALSET][i]==KEY_POINTC) Keys[DECIMALSET][i]=KEY_POINTP;
			if (Keys[COMPLEXSET][i]==KEY_POINTC) Keys[COMPLEXSET][i]=KEY_POINTP;
			if (Keys[COMPLEXSET][i]==KEY_IMAGP)  Keys[COMPLEXSET][i]=KEY_IMAGC;
		}
	}
}

// Another fudge to make hyperbolic functions possible.

void HyperbolicKeys(BOOL bIsHyp)
{
	static BOOL bOldIsHyp = FALSE;
	static int HYP_KEYS[]    = {KEY_NONHYP,KEY_SINH,KEY_COSH,KEY_TANH,KEY_ASINH,KEY_ACOSH,KEY_ATANH};
	static int NONHYP_KEYS[] = {KEY_HYP,   KEY_SIN, KEY_COS, KEY_TAN, KEY_ASIN, KEY_ACOS, KEY_ATAN};
	int iMode = RADIX_LIN;
	int i,j;
	
	if (!Pane1 || bIsHyp == bOldIsHyp) return;
	bOldIsHyp = bIsHyp;
	
	if (bIsHyp) {
		for (i=0;i<nKeyRows*nKeyCols;i++) {
			for (j=0;j<sizeof(HYP_KEYS)/sizeof(HYP_KEYS[0]);j++) {
				if (Keys[DECIMALSET][i]==NONHYP_KEYS[j]) Keys[DECIMALSET][i]=HYP_KEYS[j];
				if (Keys[COMPLEXSET][i]==NONHYP_KEYS[j]) Keys[COMPLEXSET][i]=HYP_KEYS[j];
			}
		}
	} else {
		for (i=0;i<nKeyRows*nKeyCols;i++) {
			for (j=0;j<sizeof(HYP_KEYS)/sizeof(HYP_KEYS[0]);j++) {
				if (Keys[DECIMALSET][i]==HYP_KEYS[j]) Keys[DECIMALSET][i]=NONHYP_KEYS[j];
				if (Keys[COMPLEXSET][i]==HYP_KEYS[j]) Keys[COMPLEXSET][i]=NONHYP_KEYS[j];
			}
		}
	}
	for (i=0;i<nKeyRows*nKeyCols;i++) {
		for (j=0;j<sizeof(HYP_KEYS)/sizeof(HYP_KEYS[0]);j++) {
			if (Keys[iMode][i]==HYP_KEYS[j] || Keys[iMode][i]==NONHYP_KEYS[j]) InvalidateRect(hwndKey[i],0,FALSE);
		}
	}
}

// Yet a couple to display current selection for Quick convert/constant

void QConvText(char *text)
{
	static char qconvmgs[128];
	lstrcpy(qconvmgs,text);
	lstrcat(qconvmgs," [Shift-V]");
	Key[KEY_QCONV].tip = qconvmgs;
}

void QConstText(char *text)
{
	static char qconstmgs[128];
	lstrcpy(qconstmgs,text);
	lstrcat(qconstmgs," [Shift-K]");
	Key[KEY_QCONS].tip = qconstmgs;
}

LRESULT CALLBACK KeyProc(HWND hwnd,UINT wMsg,WPARAM wParam,LPARAM lParam)
{
	static HWND hwndButtonDown = NULL; // the onscreen button
	static BOOL bMouseDown = FALSE;	// TRUE when mousebutton down
	static BOOL editing = FALSE;
	RECT rect;
	POINT pt;
	HMENU hpopup = NULL;
	WORD wID;
	
	int iMode;
	int iWindow;
	key *k;
	char menuitem[128];
	int m;

	// find which window this is, and which key, based on mode
	iWindow = GetWindowLong(hwnd,GWL_ID);
	iMode = RADIX_LIN;
	k = Key+Keys[iMode][iWindow];
	
	// Relay to XCALC ToolTip window unless empty
	if (k->id) {
		GetClientRect(hwnd,&rect);
		rect.left += Border;
		rect.top += Border;
		rect.right -= Border;
		rect.bottom -= Border;
		if (k->op[0] == _AS && k->op[1] == AS_UNDO) {
			// Add undo information to undo tooltip
			char sUndoDesc[TOOLTIP_MAXLENGTH];
			lstrcpy(sUndoDesc,"Undo ");
			lstrcat(sUndoDesc,UndoMsg[UndoIndex>0?UndoInfo[UndoIndex-1].UndoType:0]);
			lstrcat(sUndoDesc,"\n[Ctrl-z or Alt-Backspace]");
			ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,sUndoDesc);
		}
		else {
			ToolTipEvent(hwnd,wMsg,wParam,lParam,&rect,k->tip);
		}
	}

	switch (wMsg) {
	case WM_LBUTTONDOWN:
		KeyLButtonDown(hwnd,lParam,&hwndButtonDown,&bMouseDown);
		break;
		
	case WM_LBUTTONUP:
		if (KeyLButtonUp(hwnd,lParam,&hwndButtonDown,&bMouseDown)) {
			// fudge here for Undo key (and other keys to send AS_xxxx):
			if (k->op[0] == _AS)
				PostMessage(GetParent(hwnd),WM_COMMAND,k->op[1],0);
			else if (k->op[0]) {
				PostMessage(GetParent(hwnd),WM_KEYDOWN,k->op[0],0);
				if (k->op[1]) {
					PostMessage(GetParent(hwnd),WM_KEYDOWN,k->op[1],0);
					PostMessage(GetParent(hwnd),WM_KEYUP,k->op[1],0);
				}
				PostMessage(GetParent(hwnd),WM_KEYUP,k->op[0],0);
			}
			else if (k->op[1]) {
				PostMessage(GetParent(hwnd),WM_CHAR,k->op[1],0);
			}
		}
		break;
		
	case WM_MOUSEMOVE:
		// All ToolTip functionality is done in ToolTipEvent (above).
		if (bMouseDown) {
			KeyMouseDownMove(hwnd,lParam,&hwndButtonDown);
		}
		break;

	case WM_RBUTTONDOWN:
		if (AllowEdit) {
			KeyToEdit = k->id;
			ModeToEdit = iMode;
			WindowToEdit = iWindow;
			HwndToEdit = hwnd;
			GetClientRect(hwnd,&RectToEdit);
			ClientToScreen(hwnd,(POINT*)&RectToEdit.left);
			ClientToScreen(hwnd,(POINT*)&RectToEdit.right);
			pt.x = LOWORD(lParam);
			pt.y = HIWORD(lParam);
			ClientToScreen(hwnd, &pt);
			// Create popup menu
			hpopup = CreatePopupMenu();
			/* Requires quite a bit of drag/drop work!
			AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_EDITKEYS,"Edit keys...");
			*/
			AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_EDIT,"Edit this key...");
			if (nKeyCols==10 || nKeyCols==11 || nKeyCols==12) {
				AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_RESTORE,"Restore this key");
				lstrcpy(menuitem,"Restore all ");
				lstrcat(menuitem,ModeNames[RADIX_LIN]);
				lstrcat(menuitem," keys");
				AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_RESTOREALL,menuitem);
				AppendMenu(hpopup,MF_ENABLED|MF_STRING,IDM_RESTOREALLMODES,"Restore all keys");
			}
			editing = TRUE;
			InvalidateRect(HwndToEdit,NULL,FALSE);
			TrackPopupMenu(hpopup,TPM_LEFTALIGN|TPM_RIGHTBUTTON,pt.x,pt.y,0,hwnd,NULL);
			editing = FALSE;
			InvalidateRect(HwndToEdit,NULL,FALSE);
		}
		break;

	case WM_COMMAND:
		wID = GET_WM_COMMAND_ID(wParam,lParam);
		switch (wID) {
		/* Later? - requres drag&drop
		case IDM_EDITKEYS:
			PostMessage(hwndXCALC,PM_EDITKEYS,0,0);
			break;
		*/
		case IDM_EDITPROGRAM:
			PostMessage(hwndXCALC,PM_EDITPROGRAM,0,0);
			break;
		case IDM_EDIT:
			editing = TRUE;
			InvalidateRect(HwndToEdit,NULL,FALSE);
			if (DialogBox(hInst,"EDITKEY",hwndXCALC,EditKeyDlgProc)) { // true=change key
				Keys[ModeToEdit][WindowToEdit] = KeyToEdit;
			}
			editing = FALSE;
			KeyboardDecimal(k_useComma);
			HyperbolicKeys(FALSE);
			InvalidateRect(HwndToEdit,NULL,FALSE);
			break;
		case IDM_RESTORE:
			if (WindowToEdit<NDFLTKEYS) {
				const int (*dflt)[NRADIX][NDFLTKEYS] = dfltKeys();
				Keys[ModeToEdit][WindowToEdit] = (*dflt)[ModeToEdit][WindowToEdit];
				KeyboardDecimal(k_useComma);
				HyperbolicKeys(FALSE);
				InvalidateRect(HwndToEdit,NULL,FALSE);
			}
			break;
		case IDM_RESTOREALL:
			SetDefaultKeyboard(ModeToEdit);
			KeyboardDecimal(k_useComma);
			HyperbolicKeys(FALSE);
			InvalidateRect(GetParent(hwnd),NULL,FALSE);
			break;
		case IDM_RESTOREALLMODES:
			for (m=0;m<NRADIX;m++) SetDefaultKeyboard(m);
			KeyboardDecimal(k_useComma);
			HyperbolicKeys(FALSE);
			InvalidateRect(GetParent(hwnd),NULL,FALSE);
			break;
		}
		if (hpopup) DestroyMenu(hpopup);
		break;

	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			RECT rect;
			BeginPaint(hwnd,&ps);
			GetClientRect(hwnd,&rect);
			FillRect(ps.hdc,&rect,hBackBrush);
			// Reduce rect to disallow more border drawing
			rect.top+=Border+1;
			rect.left+=Border+1;
			rect.bottom-=Border;
			rect.right-=Border;
			//3d look and text if key is in use
			if (k->caption[0]) {
				DrawKey(ps.hdc,&rect,hwndButtonDown==hwnd,GetColour(RGBKEYCAPS),
					k->caption,
					k->symbolPos,
					lstrlen(k->caption));
			}
			if (editing && hwnd==HwndToEdit) {
				HPEN *oldPen = SelectObject(ps.hdc,hSelectedPen);
				HBRUSH *oldBrush = SelectObject(ps.hdc,GetStockObject(NULL_BRUSH));
				Rectangle(ps.hdc,rect.left,rect.top,rect.right,rect.bottom);
				Rectangle(ps.hdc,rect.left-1,rect.top-1,rect.right+1,rect.bottom+1);
				SelectObject(ps.hdc,oldPen);
				SelectObject(ps.hdc,oldBrush);
			}
			EndPaint(hwnd,&ps);
			break;
		}
		
	default:
		return DefWindowProc(hwnd, wMsg, wParam, lParam);
	}
	return 0L;
}

#ifdef __BORLANDC__
#pragma option -d-
#endif
